# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import torch
import torch.nn as nn
from torch.nn import functional as F
import numpy as np
import math

class JointHeatmapLoss(nn.Module):
    def __ini__(self):
        super(JointHeatmapLoss, self).__init__()

    def forward(self, joint_out, joint_gt, joint_valid):
        loss = (joint_out - joint_gt)**2 * joint_valid[:,:,None,None,None]
        return loss

class HandTypeLoss(nn.Module):
    def __init__(self):
        super(HandTypeLoss, self).__init__()

    def forward(self, hand_type_out, hand_type_gt, hand_type_valid):
        loss = F.binary_cross_entropy(hand_type_out, hand_type_gt, reduction='none')
        loss = loss.mean(1)
        loss = loss * hand_type_valid

        return loss

class RelRootDepthLoss(nn.Module):
    def __init__(self):
        super(RelRootDepthLoss, self).__init__()

    def forward(self, root_depth_out, root_depth_gt, root_valid):
        loss = torch.abs(root_depth_out - root_depth_gt) * root_valid
        return loss




# 优化器配置
optimizer = dict(type='SGD', lr=0.05,  weight_decay=5e-4, momentum=0.01, nesterov=True)
optimizer_config = dict(grad_clip=None)

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
)

# 日志配置
log_config = dict(
    interval=50,    
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 模型配置
model = dict(
    type='ImageClassifier',
    backbone=dict(
        type='WideResNet',
        num_classes=10,     # TODO 28
        depth=28,
        widen_factor=2,     # TODO 8
        dropout=0,
        dense_dropout=0.2          
    ),
    head=dict(
        type='ClsHead'
    )
)

# checkpoint配置
checkpoint_config = dict(interval=1, out_dir='./output/')       

# 数据配置
data=dict(
    train=dict( 
        type="Cifar10",
        train_or_test='train',
        dir='./cifar10-dataset',
        pipeline=[    
            dict(type='INumpyToPIL', keys=['image'], mode='RGB'),
            dict(type='RandomHorizontalFlip', keys=['image']),
            dict(type='RandomCrop', size=(32,32), padding=int(32*0.125), fill=128, padding_mode='constant', keys=['image']), 
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def={'fields': ['image', 'label']}     
    ),
    train_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=True,
        shuffle=True,
    ),
    val=dict(
        type="Cifar10",
        train_or_test='test',
        dir='./cifar10-dataset',
        pipeline=[    
            dict(type='INumpyToPIL', keys=['image'], mode='RGB'),                  
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def={'fields': ['image', 'label']}  
    ),
    val_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=False,
        shuffle=False,
    ),
    test=dict(
        type="Cifar10",
        train_or_test='test',
        dir='./cifar10-dataset',
        pipeline=[    
            dict(type='INumpyToPIL', keys=['image'], mode='RGB'),                  
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def={'fields': ['image', 'label']}  
    ),
    test_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=False,
        shuffle=False,
    ),    
)

# 评估方案配置
evaluation=dict(out_dir='./output/', interval=1, metric=dict(type='AccuracyEval'), save_best='top_1')

# 导出配置
export=dict(
    input_shape_list = [[1,3,32,32]],
    input_name_list=["image"],
    output_name_list=["pred"]
)

max_epochs = 1500


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F

from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class Hint(nn.Module):
	'''
	FitNets: Hints for Thin Deep Nets
	https://arxiv.org/pdf/1412.6550.pdf
	'''
	def __init__(self):
		super(Hint, self).__init__()

	def forward(self, fm_s, fm_t):
		loss = F.mse_loss(fm_s, fm_t)

		return loss


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES



'''
AT with sum of absolute values with power p
'''

@DISTILL_LOSSES.register_module()
class AT(nn.Module):
	'''
	Paying More Attention to Attention: Improving the Performance of Convolutional
	Neural Netkworks wia Attention Transfer
	https://arxiv.org/pdf/1612.03928.pdf
	'''
	def __init__(self, p):
		super(AT, self).__init__()
		self.p = p

	def forward(self, fm_s, fm_t):
		loss = F.mse_loss(self.attention_map(fm_s), self.attention_map(fm_t))

		return loss

	def attention_map(self, fm, eps=1e-6):
		am = torch.pow(torch.abs(fm), self.p)
		am = torch.sum(am, dim=1, keepdim=True)
		norm = torch.norm(am, dim=(2,3), keepdim=True)
		am = torch.div(am, norm+eps)

		return am


# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 13:18
# @File    : util.py
# @Author  : jian<jian@mltalker.com>

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import argparse
import importlib
import os
import sys

def load_extmodule(ext_module_file):
    ext_module_dir = os.path.dirname(ext_module_file)
    key_model = ext_module_file
    dot_pos = key_model.rfind(".")
    if dot_pos != -1:
        key_model = key_model[0:dot_pos]

    sys.path.append(ext_module_dir)
    key_model = os.path.normpath(key_model)
    key_model = os.path.relpath(key_model, ext_module_dir)
    key_model = key_model.replace('/', '.')
    if key_model.startswith('.'):
        key_model = key_model[1:]
    
    importlib.import_module(key_model)


import torch
import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.models.builder import MODELS
from antgo.framework.helper.runner import BaseModule
from antgo.framework.helper.utils import Registry, build_from_cfg

@MODELS.register_module()
class ACModule(BaseModule):
    def __init__(self, model, test_cfg, train_cfg=None, init_cfg=None):
        super().__init__(init_cfg)
        self.model = build_from_cfg(model, MODELS)

        # feature_from, uncertainty_from, bayesian_from
        self.test_cfg = test_cfg
 
    def forward(self, *args, **kwargs):
        out = {}

        feature_extract_proxy = None
        if self.test_cfg.get('feature_config', None) is not None:
            feature_extract_proxy = getattr(getattr(self.model, self.test_cfg.feature_config.from_name), 'forward', None)
            from_index = self.test_cfg.feature_config.from_index
            def feature_extract_func(x):
                x = feature_extract_proxy(x)
                feature = x
                if isinstance(x, tuple):
                    feature = x[from_index]
                batch_size = feature.shape[0]
                out.update({
                    'feature': feature.view(batch_size, -1)
                })
                return x

            setattr(getattr(self.model, self.test_cfg.feature_config.from_name), 'forward', feature_extract_func)

        uncertainty_extract_proxy = None
        if self.test_cfg.get('uncertainty_config', None) is not None:
            uncertainty_extract_proxy = getattr(self.model, self.test_cfg.uncertainty_config.from_name, None)
            with_sigmoid = self.test_cfg.uncertainty_config.with_sigmoid
            from_index = self.test_cfg.feature_config.from_index
            def uncertainty_extract_func(x):
                x = uncertainty_extract_proxy(x)
                probability = x
                if isinstance(x, tuple):
                    probability = x[from_index]
                                
                if with_sigmoid:
                    probability = probability.sigmoid()
                                    
                out.update({
                    'uncertainty': probability
                })
                return x
            setattr(self.model, self.test_cfg.uncertainty_config.from_name, uncertainty_extract_func)

        bayesian_proxy = None
        if self.test_cfg.get('bayesian_config', None) is not None:
            bayesian_proxy = getattr(self.model, self.test_cfg.bayesian_config.from_name, None)
            p = self.test_cfg.bayesian_config.p
            def bayesian_func(x):
                x = bayesian_proxy(x)
                x = F.dropout(x, p,training=True)
                return x
            setattr(self.model, self.test_cfg.bayesian_config.from_name, bayesian_func)

        result = self.model(*args, **kwargs)

        # 恢复
        if feature_extract_proxy is not None:
            setattr(getattr(self.model, self.test_cfg.feature_config.from_name), 'forward', feature_extract_proxy)
        if uncertainty_extract_proxy is not None:
            setattr(self.model, self.test_cfg.uncertainty_config.from_name, uncertainty_extract_proxy)
        if bayesian_proxy is not None:
            setattr(self.model, self.test_cfg.bayesian_config.from_name, bayesian_proxy)

        out.update(result)
        return out



import torch
from torch.utils.data import Dataset


class DummyModelWarp(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
    
    def forward(self, image):
        # forward_dumpy is no redundant logic 
        return self.model.forward_dummy(image)

class UnlabeledDatasetWrapper(Dataset):
    def __init__(self, dataset, key):
        self.dataset = dataset
        self.key = key

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        images = self.dataset[index][self.key]
        return images

class LabeledDatasetWrapper(Dataset):
    def __init__(self, dataset, key):
        self.dataset = dataset
        self.key = key

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        images = self.dataset[index][self.key]
        return images, {}



"""easy dict"""


class edict(dict):
    """a dict class which could be accessed by attribute"""

    def __setattr__(self, key, value):
        super(edict, self).__setitem__(key, value)

    def __getattr__(self, key):
        try:
            return super(edict, self).__getitem__(key)
        except KeyError:
            raise AttributeError(key)

    def __delattr__(self, key):
        super(edict, self).__delitem__(key)



# -*- coding:UTF-8 -*-
# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import hashlib
import requests
import threading
import uuid
import json

stats_api = "http://paddlepaddle.org.cn/paddlehub/stat"


def string(param):
    """ 生成字符串。
    """
    return "\'{}\'".format(param)


def check_version():
    version = paddle.__version__
    v0, v1, v2 = version.split('.')
    if not ((v0 == '0' and v1 == '0' and v2 == '0') or
            (int(v0) >= 2 and int(v1) >= 1)):
        return False
    else:
        return True


def _md5(text: str):
    '''Calculate the md5 value of the input text.'''
    md5code = hashlib.md5(text.encode())
    return md5code.hexdigest()


class ConverterCheck(threading.Thread):
    """
    Count the number of calls to model convertion
    """

    def __init__(self,
                 task="ONNX",
                 time_info=None,
                 convert_state=None,
                 lite_state=None,
                 extra_info=None):
        threading.Thread.__init__(self)
        self._task = task
        self._version = x2paddle.__version__
        self._convert_state = convert_state
        self._lite_state = lite_state
        self._extra_info = extra_info
        self._convert_id = _md5(str(uuid.uuid1())[-12:]) + "-" + str(time_info)

    def run(self):
        params = {
            'task': self._task,
            'x2paddle_version': self._version,
            'paddle_version': paddle.__version__,
            'from': 'x2paddle'
        }
        extra = {
            'convert_state': self._convert_state,
            'convert_id': self._convert_id,
        }
        if self._lite_state is not None:
            extra.update({'lite_state': self._lite_state})
        if self._extra_info is not None:
            extra.update(self._extra_info)

        params.update({"extra": json.dumps(extra)})
        try:
            requests.get(stats_api, params, timeout=2)
        except Exception:
            pass

        return


class PaddleDtypes():
    def __init__(self, is_new_version=True):
        if is_new_version:
            self.t_float16 = paddle.float16
            self.t_float32 = paddle.float32
            self.t_float64 = paddle.float64
            self.t_uint8 = paddle.uint8
            self.t_int8 = paddle.int8
            self.t_int16 = paddle.int16
            self.t_int32 = paddle.int32
            self.t_int64 = paddle.int64
            self.t_bool = paddle.bool
        else:
            self.t_float16 = "paddle.fluid.core.VarDesc.VarType.FP16"
            self.t_float32 = "paddle.fluid.core.VarDesc.VarType.FP32"
            self.t_float64 = "paddle.fluid.core.VarDesc.VarType.FP64"
            self.t_uint8 = "paddle.fluid.core.VarDesc.VarType.UINT8"
            self.t_int8 = "paddle.fluid.core.VarDesc.VarType.INT8"
            self.t_int16 = "paddle.fluid.core.VarDesc.VarType.INT16"
            self.t_int32 = "paddle.fluid.core.VarDesc.VarType.INT32"
            self.t_int64 = "paddle.fluid.core.VarDesc.VarType.INT64"
            self.t_bool = "paddle.fluid.core.VarDesc.VarType.BOOL"


is_new_version = check_version()
paddle_dtypes = PaddleDtypes(is_new_version)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 22:03
# @File    : dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from typing import Union
from pathlib import Path

from antgo.pipeline.functional.entity import Entity


class DatasetMixin:
    """
    Mixin for dealing with dataset
    """

    # pylint: disable=import-outside-toplevel
    @classmethod
    def from_glob(cls, *args):  # pragma: no cover
        """
        generate a file list with `pattern`
        """
        from glob import glob
        files = []
        for path in args:
            files.extend(glob(path))
        if len(files) == 0:
            raise FileNotFoundError(f'There is no files with {args}.')
        return cls(files)

    @classmethod
    def placeholder(cls, *args):
        signal_set = [
            'EAGLEEYE_SIGNAL_RGB_IMAGE',
            'EAGLEEYE_SIGNAL_BGR_IMAGE',
            'EAGLEEYE_SIGNAL_RGBA_IMAGE',
            'EAGLEEYE_SIGNAL_BGRA_IMAGE',
            'EAGLEEYE_SIGNAL_GRAY_IMAGE',
            'EAGLEEYE_SIGNAL_MASK',
            'EAGLEEYE_SIGNAL_DET',
            'EAGLEEYE_SIGNAL_DET_EXT',
            'EAGLEEYE_SIGNAL_TRACKING',
            'EAGLEEYE_SIGNAL_POS_2D',
            'EAGLEEYE_SIGNAL_POS_3D',
            'EAGLEEYE_SIGNAL_LANDMARK',
            'EAGLEEYE_SIGNAL_CLS',
            'EAGLEEYE_SIGNAL_STATE',
            'EAGLEEYE_SIGNAL_SWITCH',
            'EAGLEEYE_SIGNAL_RECT',
            'EAGLEEYE_SIGNAL_LINE',
            'EAGLEEYE_SIGNAL_POINT'
        ]
        for arg in args:
            if isinstance(arg, tuple):
                assert (arg[-1] in signal_set)
        return cls(list(args))

    @classmethod
    def read_json(cls, json_path: Union[str, Path], encoding: str = 'utf-8'):
        import json

        def inner():
            with open(json_path, 'r', encoding=encoding) as f:
                string = f.readline()
                while string:
                    data = json.loads(string)
                    string = f.readline()
                    yield Entity(**data)

        return cls(inner())

    def random_sample(self):
        # core API already exists
        pass

    def filter_data(self):
        # core API already exists
        pass




import copy
import random
from tqdm import tqdm
import numpy as np

__all__ = ["EvolutionFinder"]


class ArchManager:
    def __init__(self):
        self.num_blocks = 20
        self.num_stages = 5
        self.kernel_sizes = [3, 5, 7]
        self.expand_ratios = [3, 4, 6]
        self.depths = [2, 3, 4]
        self.resolutions = [160, 176, 192, 208, 224]

    def random_sample(self):
        sample = {}
        d = []
        e = []
        ks = []
        for i in range(self.num_stages):
            d.append(random.choice(self.depths))

        for i in range(self.num_blocks):
            e.append(random.choice(self.expand_ratios))
            ks.append(random.choice(self.kernel_sizes))

        sample = {
            "wid": None,
            "ks": ks,
            "e": e,
            "d": d,
            "r": [random.choice(self.resolutions)],
        }

        return sample

    def random_resample(self, sample, i):
        assert i >= 0 and i < self.num_blocks
        sample["ks"][i] = random.choice(self.kernel_sizes)
        sample["e"][i] = random.choice(self.expand_ratios)

    def random_resample_depth(self, sample, i):
        assert i >= 0 and i < self.num_stages
        sample["d"][i] = random.choice(self.depths)

    def random_resample_resolution(self, sample):
        sample["r"][0] = random.choice(self.resolutions)


class EvolutionFinder:
    valid_constraint_range = {
        "flops": [150, 600],
        "note10": [15, 60],
    }

    def __init__(
        self,
        constraint_type,
        efficiency_constraint,
        efficiency_predictor,
        accuracy_predictor,
        **kwargs
    ):
        self.constraint_type = constraint_type
        if not constraint_type in self.valid_constraint_range.keys():
            self.invite_reset_constraint_type()
        self.efficiency_constraint = efficiency_constraint
        if not (
            efficiency_constraint <= self.valid_constraint_range[constraint_type][1]
            and efficiency_constraint >= self.valid_constraint_range[constraint_type][0]
        ):
            self.invite_reset_constraint()

        self.efficiency_predictor = efficiency_predictor
        self.accuracy_predictor = accuracy_predictor
        self.arch_manager = ArchManager()
        self.num_blocks = self.arch_manager.num_blocks
        self.num_stages = self.arch_manager.num_stages

        self.mutate_prob = kwargs.get("mutate_prob", 0.1)
        self.population_size = kwargs.get("population_size", 100)
        self.max_time_budget = kwargs.get("max_time_budget", 500)
        self.parent_ratio = kwargs.get("parent_ratio", 0.25)
        self.mutation_ratio = kwargs.get("mutation_ratio", 0.5)

    def invite_reset_constraint_type(self):
        print(
            "Invalid constraint type! Please input one of:",
            list(self.valid_constraint_range.keys()),
        )
        new_type = input()
        while new_type not in self.valid_constraint_range.keys():
            print(
                "Invalid constraint type! Please input one of:",
                list(self.valid_constraint_range.keys()),
            )
            new_type = input()
        self.constraint_type = new_type

    def invite_reset_constraint(self):
        print(
            "Invalid constraint_value! Please input an integer in interval: [%d, %d]!"
            % (
                self.valid_constraint_range[self.constraint_type][0],
                self.valid_constraint_range[self.constraint_type][1],
            )
        )

        new_cons = input()
        while (
            (not new_cons.isdigit())
            or (int(new_cons) > self.valid_constraint_range[self.constraint_type][1])
            or (int(new_cons) < self.valid_constraint_range[self.constraint_type][0])
        ):
            print(
                "Invalid constraint_value! Please input an integer in interval: [%d, %d]!"
                % (
                    self.valid_constraint_range[self.constraint_type][0],
                    self.valid_constraint_range[self.constraint_type][1],
                )
            )
            new_cons = input()
        new_cons = int(new_cons)
        self.efficiency_constraint = new_cons

    def set_efficiency_constraint(self, new_constraint):
        self.efficiency_constraint = new_constraint

    def random_sample(self):
        constraint = self.efficiency_constraint
        while True:
            sample = self.arch_manager.random_sample()
            efficiency = self.efficiency_predictor.predict_efficiency(sample)
            if efficiency <= constraint:
                return sample, efficiency

    def mutate_sample(self, sample):
        constraint = self.efficiency_constraint
        while True:
            new_sample = copy.deepcopy(sample)

            if random.random() < self.mutate_prob:
                self.arch_manager.random_resample_resolution(new_sample)

            for i in range(self.num_blocks):
                if random.random() < self.mutate_prob:
                    self.arch_manager.random_resample(new_sample, i)

            for i in range(self.num_stages):
                if random.random() < self.mutate_prob:
                    self.arch_manager.random_resample_depth(new_sample, i)

            efficiency = self.efficiency_predictor.predict_efficiency(new_sample)
            if efficiency <= constraint:
                return new_sample, efficiency

    def crossover_sample(self, sample1, sample2):
        constraint = self.efficiency_constraint
        while True:
            new_sample = copy.deepcopy(sample1)
            for key in new_sample.keys():
                if not isinstance(new_sample[key], list):
                    continue
                for i in range(len(new_sample[key])):
                    new_sample[key][i] = random.choice(
                        [sample1[key][i], sample2[key][i]]
                    )

            efficiency = self.efficiency_predictor.predict_efficiency(new_sample)
            if efficiency <= constraint:
                return new_sample, efficiency

    def run_evolution_search(self, verbose=False):
        """Run a single roll-out of regularized evolution to a fixed time budget."""
        max_time_budget = self.max_time_budget
        population_size = self.population_size
        mutation_numbers = int(round(self.mutation_ratio * population_size))
        parents_size = int(round(self.parent_ratio * population_size))
        constraint = self.efficiency_constraint

        best_valids = [-100]
        population = []  # (validation, sample, latency) tuples
        child_pool = []
        efficiency_pool = []
        best_info = None
        if verbose:
            print("Generate random population...")
        for _ in range(population_size):
            sample, efficiency = self.random_sample()
            child_pool.append(sample)
            efficiency_pool.append(efficiency)

        accs = self.accuracy_predictor.predict_accuracy(child_pool)
        for i in range(population_size):
            population.append((accs[i].item(), child_pool[i], efficiency_pool[i]))

        others_info = []
        if verbose:
            print("Start Evolution...")
        # After the population is seeded, proceed with evolving the population.
        for iter in tqdm(
            range(max_time_budget),
            desc="Searching with %s constraint (%s)"
            % (self.constraint_type, self.efficiency_constraint),
        ):
            parents = sorted(population, key=lambda x: x[0])[::-1][:parents_size]
            acc = parents[0][0]
            if verbose:
                print("Iter: {} Acc: {}".format(iter - 1, parents[0][0]))

            if acc > best_valids[-1]:
                best_valids.append(acc)
                best_info = parents[0]
            else:
                best_valids.append(best_valids[-1])

            population = parents
            child_pool = []
            efficiency_pool = []

            for i in range(mutation_numbers):
                par_sample = population[np.random.randint(parents_size)][1]
                # Mutate
                new_sample, efficiency = self.mutate_sample(par_sample)
                child_pool.append(new_sample)
                efficiency_pool.append(efficiency)

            for i in range(population_size - mutation_numbers):
                par_sample1 = population[np.random.randint(parents_size)][1]
                par_sample2 = population[np.random.randint(parents_size)][1]
                # Crossover
                new_sample, efficiency = self.crossover_sample(par_sample1, par_sample2)
                child_pool.append(new_sample)
                efficiency_pool.append(efficiency)

            accs = self.accuracy_predictor.predict_accuracy(child_pool)
            for i in range(population_size):
                population.append((accs[i].item(), child_pool[i], efficiency_pool[i]))

            others_info.append(best_info)

        return best_valids, best_info, others_info



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from ..builder import BBOX_SAMPLERS
from ..transforms import bbox2roi
from .base_sampler import BaseSampler


@BBOX_SAMPLERS.register_module()
class OHEMSampler(BaseSampler):
    r"""Online Hard Example Mining Sampler described in `Training Region-based
    Object Detectors with Online Hard Example Mining
    <https://arxiv.org/abs/1604.03540>`_.
    """

    def __init__(self,
                 num,
                 pos_fraction,
                 context,
                 neg_pos_ub=-1,
                 add_gt_as_proposals=True,
                 loss_key='loss_cls',
                 **kwargs):
        super(OHEMSampler, self).__init__(num, pos_fraction, neg_pos_ub,
                                          add_gt_as_proposals)
        self.context = context
        if not hasattr(self.context, 'num_stages'):
            self.bbox_head = self.context.bbox_head
        else:
            self.bbox_head = self.context.bbox_head[self.context.current_stage]

        self.loss_key = loss_key

    def hard_mining(self, inds, num_expected, bboxes, labels, feats):
        with torch.no_grad():
            rois = bbox2roi([bboxes])
            if not hasattr(self.context, 'num_stages'):
                bbox_results = self.context._bbox_forward(feats, rois)
            else:
                bbox_results = self.context._bbox_forward(
                    self.context.current_stage, feats, rois)
            cls_score = bbox_results['cls_score']
            loss = self.bbox_head.loss(
                cls_score=cls_score,
                bbox_pred=None,
                rois=rois,
                labels=labels,
                label_weights=cls_score.new_ones(cls_score.size(0)),
                bbox_targets=None,
                bbox_weights=None,
                reduction_override='none')[self.loss_key]
            _, topk_loss_inds = loss.topk(num_expected)
        return inds[topk_loss_inds]

    def _sample_pos(self,
                    assign_result,
                    num_expected,
                    bboxes=None,
                    feats=None,
                    **kwargs):
        """Sample positive boxes.

        Args:
            assign_result (:obj:`AssignResult`): Assigned results
            num_expected (int): Number of expected positive samples
            bboxes (torch.Tensor, optional): Boxes. Defaults to None.
            feats (list[torch.Tensor], optional): Multi-level features.
                Defaults to None.

        Returns:
            torch.Tensor: Indices  of positive samples
        """
        # Sample some hard positive samples
        pos_inds = torch.nonzero(assign_result.gt_inds > 0, as_tuple=False)
        if pos_inds.numel() != 0:
            pos_inds = pos_inds.squeeze(1)
        if pos_inds.numel() <= num_expected:
            return pos_inds
        else:
            return self.hard_mining(pos_inds, num_expected, bboxes[pos_inds],
                                    assign_result.labels[pos_inds], feats)

    def _sample_neg(self,
                    assign_result,
                    num_expected,
                    bboxes=None,
                    feats=None,
                    **kwargs):
        """Sample negative boxes.

        Args:
            assign_result (:obj:`AssignResult`): Assigned results
            num_expected (int): Number of expected negative samples
            bboxes (torch.Tensor, optional): Boxes. Defaults to None.
            feats (list[torch.Tensor], optional): Multi-level features.
                Defaults to None.

        Returns:
            torch.Tensor: Indices  of negative samples
        """
        # Sample some hard negative samples
        neg_inds = torch.nonzero(assign_result.gt_inds == 0, as_tuple=False)
        if neg_inds.numel() != 0:
            neg_inds = neg_inds.squeeze(1)
        if len(neg_inds) <= num_expected:
            return neg_inds
        else:
            neg_labels = assign_result.labels.new_empty(
                neg_inds.size(0)).fill_(self.bbox_head.num_classes)
            return self.hard_mining(neg_inds, num_expected, bboxes[neg_inds],
                                    neg_labels, feats)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/25 21:27
# @File    : restoration.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os

import mmcv
import torch

from mmedit.apis import init_model
from mmedit.core import tensor2img
from mmedit.utils import modify_args
from mmcv.parallel import collate, scatter

from mmedit.datasets.pipelines import Compose



def _restoration_inference(model, img, ref=None):
  """Inference image with the model.

  Args:
      model (nn.Module): The loaded model.
      img (str): File path of input image.
      ref (str | None): File path of reference image. Default: None.

  Returns:
      Tensor: The predicted restoration result.
  """
  cfg = model.cfg
  device = next(model.parameters()).device  # model device
  # remove gt from test_pipeline
  keys_to_remove = ['gt', 'gt_path', 'lq_path']
  for key in keys_to_remove:
    for pipeline in list(cfg.test_pipeline):
      if 'key' in pipeline and key == pipeline['key']:
        cfg.test_pipeline.remove(pipeline)
      if 'keys' in pipeline and key in pipeline['keys']:
        pipeline['keys'].remove(key)
        if len(pipeline['keys']) == 0:
          cfg.test_pipeline.remove(pipeline)
      if 'meta_keys' in pipeline and key in pipeline['meta_keys']:
        pipeline['meta_keys'].remove(key)

  for pipeline in list(cfg.test_pipeline):
    if pipeline['type'] == 'LoadImageFromFile':
      cfg.test_pipeline.remove(pipeline)

  # build the data pipeline
  test_pipeline = Compose(cfg.test_pipeline)
  # prepare data
  if ref:  # Ref-SR
    data = dict(lq=img, ref=ref)
  else:  # SISR
    data = dict(lq=img)
  data = test_pipeline(data)
  data = collate([data], samples_per_gpu=1)
  if 'cuda' in str(device):
    data = scatter(data, [device])[0]
  # forward the model
  with torch.no_grad():
    result = model(test_mode=True, **data)

  return result['output']


class Restoration(object):
  def __init__(self, config_file, checkpoint_file, device='cpu'):
    self.model = init_model(config_file, checkpoint_file, device=device)

  def __call__(self, *args, **kwargs):
    img = None
    ref = None
    if len(args) == 2:
      img, ref = args
    else:
      img = args[0]

    result = _restoration_inference(self.model, img, ref)
    return result


from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.framework.helper.parallel.utils import is_module_wrapper
from antgo.framework.helper.models.proxy_module import ProxyModule
from antgo.framework.helper.runner.test import single_gpu_test

import os
import os.path as osp
import math
import time
import glob
import abc
from torch.utils.data import DataLoader, dataloader
import torch.optim
import torchvision.transforms as transforms
from torch.nn.parallel.data_parallel import DataParallel
from antgo.framework.helper.utils.config import Config
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import (build_dataset, build_dataloader, build_kv_dataloader, build_iter_dataloader)
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.utils import get_logger
from antgo.framework.helper.runner import *
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.runner.builder import *
from antgo.framework.helper.utils.setup_env import *
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.utils import *
from antgo.framework.helper.models.dummy_module import *
from antgo.framework.helper.models.proxy_module import *
from antgo.framework.helper.models.distillation import *
from antgo.framework.helper.runner.hooks.hook import *
from antgo.framework.helper.utils import Config, build_from_cfg
from antgo.framework.helper.task_flag import *
import traceback
from thop import profile
import copy

import torch.distributed as dist
from contextlib import contextmanager
import json


class BaseTrainer(object):
    def __init__(self, cfg_dict, work_dir="./", device='cuda', distributed=False, diff_seed=True, deterministic=True):
        self.cfg = Config.fromstring(json.dumps(cfg_dict), '.json')

        self.data_loaders = None
        self.runner = None
        self.work_dir = work_dir
        self.train_generator = None
        self.val_dataloader = None
        self.distributed = distributed
        self.meta = {}

        # set multi-process settings
        # 如 opencv_num_threads, OMP_NUM_THREADS, MKL_NUM_THREADS
        setup_multi_processes(self.cfg)

        if self.distributed:
            init_dist(**self.cfg.get('dist_params', {}))
            # re-set gpu_ids with distributed training mode
            _, world_size = get_dist_info()
            self.cfg.gpu_ids = range(world_size)

        # set random seeds
        seed = init_random_seed(self.cfg.get('seed', 0), device=device)
        seed = seed + dist.get_rank() if diff_seed else seed
        set_random_seed(seed, deterministic=deterministic)
        self.cfg.seed = seed
        self.meta['seed'] = seed
        self.device = device

        self.submodule_optimizer = {}
        self.submodule_optimizer_config = {}

    def config_dataloader(self, with_validate=False):
        # 创建数据集
        dataset = build_dataset(self.cfg.data.train)

        # 创建数据集加载器
        train_dataloader_default_args = dict(
            samples_per_gpu=2,
            workers_per_gpu=2,
            # `num_gpus` will be ignored if distributed
            num_gpus=1,
            dist=self.distributed,
            seed=self.cfg.seed,
            runner_type='EpochBasedRunner',
            persistent_workers=True)

        train_loader_cfg = {
            **train_dataloader_default_args,
            **self.cfg.data.get('train_dataloader', {})
        }
        if getattr(dataset, 'is_kv', False):
            self.train_generator = build_kv_dataloader(dataset, **train_loader_cfg)
        elif isinstance(dataset, torch.utils.data.IterableDataset):
            self.train_generator = build_iter_dataloader(dataset, **train_loader_cfg)
        else:
            self.train_generator = build_dataloader(dataset, **train_loader_cfg)

        if with_validate:
            val_dataloader_default_args = dict(
                samples_per_gpu=1,
                workers_per_gpu=2,
                dist=self.distributed,
                shuffle=False,
                persistent_workers=True)
            
            val_dataloader_args = {
                    **val_dataloader_default_args,
                    **self.cfg.data.get('val_dataloader', {})
            }
            val_dataset = build_dataset(self.cfg.data.val, dict(test_mode=True))
            if getattr(dataset, 'is_kv', False):
                self.val_dataloader = build_kv_dataloader(val_dataset, **val_dataloader_args)
            elif isinstance(dataset, torch.utils.data.IterableDataset):
                val_dataset.is_infinite = False
                self.val_dataloader = build_iter_dataloader(val_dataset, **val_dataloader_args)
            else:
                self.val_dataloader = build_dataloader(val_dataset, **val_dataloader_args)

    def _config_runner(self, runner_config, model, logger):
        self.runner = build_runner(
            runner_config,        # 忽略max_epochs，开发者控制最大epoch
            default_args=dict(
                model=model,
                optimizer=self.submodule_optimizer,
                work_dir=self.work_dir,
                logger=logger,
                meta=self.meta))

    def _config_training_hooks(self):
        # 仅对通用hook进行配置
        custom_hooks = self.cfg.get('custom_hooks', None)        

        if 'policy' not in self.cfg.lr_config:
            for submodule_name, lr_config in self.cfg.lr_config.items():
                lr_config.update({'name': submodule_name})
                self.runner.register_lr_hook(lr_config)
        else:
            self.runner.register_lr_hook(self.cfg.lr_config)

        self.runner.register_checkpoint_hook(self.cfg.checkpoint_config)
        self.runner.register_logger_hooks(self.cfg.log_config)
        self.runner.register_custom_hooks(custom_hooks)

    def _config_optimizer(self, model):
        if len(self.submodule_optimizer) > 0:
            return

        if is_module_wrapper(model):
            model = model.module

        for submodule_name, optimizer_dict in self.cfg.optimizer.items():
            self.submodule_optimizer[submodule_name] = build_optimizer(getattr(model, submodule_name), optimizer_dict)
            self.submodule_optimizer_config[submodule_name] = self.cfg.optimizer_config.get(submodule_name, dict())

    def config_model(self, model_builder=None, resume_from=None, load_from=None, revise_keys=[(r'^module\.', '')]):
        # prepare network
        logger = get_logger('model', log_level=self.cfg.log_level)
        logger.info("Creating model and optimizer...")

        # 构建网络
        if model_builder is not None:
            model = model_builder()
        else:
            model = build_model(self.cfg.model)

        # 模型初始化
        model.init_weights()

        if self.distributed:
            if self.cfg.get('syncBN', True):
                model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model).to(self.device)
            
            find_unused_parameters = self.cfg.get('find_unused_parameters', False)
            # Sets the `find_unused_parameters` parameter in
            # torch.nn.parallel.DistributedDataParallel
            model = build_ddp(
                model,
                self.device,
                device_ids=[int(os.environ['LOCAL_RANK'])],
                broadcast_buffers=False,
                find_unused_parameters=find_unused_parameters)
        else:
            model = build_dp(model, self.device, device_ids=self.cfg.gpu_ids)

        # config optimizer
        self._config_optimizer(model)
        
        # 自动调整单设备下的学习率到多设备下的学习率
        auto_scale_lr(self.cfg, self.distributed, logger)

        # build training strategy
        custom_runner_config = self.cfg.get('runner', dict(type='EpochBasedRunner', max_epochs=1))
        self._config_runner(custom_runner_config, model, logger)

        # config default training hooks
        self._config_training_hooks()

        # an ugly workaround to make .log and .log.json filenames the same
        self.runner.timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())

        if self.distributed:
            if isinstance(self.runner, EpochBasedRunner):
                self.runner.register_hook(DistSamplerSeedHook())

        if resume_from is not None:
            self.runner.resume(resume_from)
        elif load_from is not None:
            self.runner.load_checkpoint(load_from, revise_keys=revise_keys)

    def start_train(self, max_epochs, **kwargs):
        try:
            running_flag(self.cfg.get('root', None))
            self.runner.run([self.train_generator], [('train', max_epochs)], max_epochs)
            finish_flag(self.cfg.get('root', None))
        except Exception:
            stop_flag(self.cfg.get('root', None))
            traceback.print_exc()

    def start_eval(self, **kwargs):
        try:
            running_flag(self.cfg.get('root', None))
            self.runner.run([self.train_generator], [('val', 1)], 1)
            finish_flag(self.cfg.get('root', None))
        except:
            stop_flag(self.cfg.get('root', None))
            traceback.print_exc()



# -*- coding: UTF-8 -*-
# @Time    : 17-5-3
# @File    : objdect_task.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.measures.average_precision import *
from antgo.measures.roc_auc import *
from antgo.measures.precision_recall import *
from antgo.measures.binary_c import *
from antgo.task.task import *
from antgo.measures.base import *
from antgo.utils._bbox import bbox_overlaps
from collections import defaultdict
import antgo.utils._mask as _mask

default = {'AntVOCDet': ('VOC', 'OBJECT-DETECTION'),
           'AntROCandAUCDet': ('ROC_AUC', 'OBJECT-DETECTION'),
           'AntPRDet': ('PR', 'OBJECT-DETECTION'),
           'AntAPRFDet': ('APRF', 'OBJECT-DETECTION'),
           'AntTFTFDet': ('TFTF', 'OBJECT-DETECTION'),
           'AntCOCODet': ('COCO', 'OBJECT-DETECTION'),
           }
class AntVOCDet(AntMeasure):
  def __init__(self, task):
    super(AntVOCDet, self).__init__(task, 'VOC')
    assert(task.task_type == 'OBJECT-DETECTION')
    self.is_support_rank = True

  def eva(self, data, label):
    category_num = len(self.task.class_label)
    detection_score = [[] for _ in range(category_num)]
    detection_label = [[] for _ in range(category_num)]

    if label is not None:
      data = zip(data, label)

    # assert(len(data) == len(label))
    predict_box_total = 0

    samples_scores = []
    # 1.step positive sample is overlap > 0.5
    for predict, gt in data:
      # sample id
      # sample_id = gt['id']

      if predict is None :
        for missed_gt_bi in range(len(gt['bbox'])):
          detection_label[gt['category_id'][missed_gt_bi]].append(1)
          detection_score[gt['category_id'][missed_gt_bi]].append(-float("inf"))
        continue

      det_bbox = predict['det-bbox']
      det_score = predict['det-score']
      det_label = predict['det-label']

      # ground truth bbox and categories
      gt_bbox = np.array(gt['bbox'])
      gt_category = np.array(gt['category_id']).astype(dtype=np.int32)
      gt_bbox_num = gt_bbox.shape[0]

      # predict position, category, and score
      predict_box = np.array(det_bbox)
      predict_category = np.array(det_label).astype(dtype=np.int32)
      predict_score = np.array(det_score)
      predict_box_total += predict_score.shape[0]

      overlaps = bbox_overlaps(
          np.ascontiguousarray(predict_box, dtype=np.float),
          np.ascontiguousarray(gt_bbox, dtype=np.float))
      #overlaps = _mask.iou(predict_box.tolist(), gt_bbox.tolist(), [0 for _ in range(gt_bbox_num)])

      # distinguish positive and negative samples and scores (overlap > 0.5)
      gtm = np.ones((gt_bbox_num)) * (-1)
      for dind, d in enumerate(predict_box.tolist()):
        # information about best match so far (m=-1 -> unmatched)
        m = -1
        iou = 0.5
        for gind, g in enumerate(gt_bbox.tolist()):
          if gt_category[gind] != predict_category[dind]:
            continue

          # if this gt already matched continue
          if gtm[gind] >= 0:
            continue

          # continue to next gt unless better match made
          if overlaps[dind, gind] < iou:
            continue
          # if match successful and best so far, store appropriately
          iou = overlaps[dind, gind]
          m = gind

        # if match made store id of match for both dt and gt
        if m > -1:
          gtm[m] = dind

          # success to match
          detection_score[gt_category[m]].append(predict_score[dind])
          detection_label[gt_category[m]].append(1)

          # # record sample
          # samples_scores.append({'id': sample_id,
          #                        'score': 1,
          #                        'category': gt_category[m],
          #                        'box': gt_bbox[m].tolist(),
          #                        'index': sample_id * 100 + m})

      # process none matched det
      for dind, d in enumerate(predict_box.tolist()):
        if dind not in gtm:
          detection_score[predict_category[dind]].append(predict_score[dind])
          detection_label[predict_category[dind]].append(0)

      # process missed gt bbox
      missed_gt_bbox = [_ for _ in range(gt_bbox_num) if gtm[_] == -1]
      for missed_gt_bi in missed_gt_bbox:
        detection_label[gt_category[missed_gt_bi]].append(1)
        detection_score[gt_category[missed_gt_bi]].append(-float("inf"))

        # # record sample
        # samples_scores.append({'id': sample_id,
        #                        'score': 0,
        #                        'category': gt_category[missed_gt_bi],
        #                        'box': gt_bbox[missed_gt_bi].tolist(),
        #                        'index': sample_id * 100 + missed_gt_bi})
      #########################################################################

    # 2.step compute mean average precision
    voc_mean_map = []
    for predict, gt in zip(detection_label, detection_score):
      result = vmap(predict, gt)
      if result is None:
        result = 0.0
      voc_mean_map.append(float(result))

    # 3.step make json
    voc_map = float(np.mean(voc_mean_map))
    return {'statistic': {'name': self.name,
                          'value': [{'name': 'Mean-MAP', 'value': voc_map, 'type': 'SCALAR'},
                                    {'name': 'MAP', 'value': voc_mean_map, 'type': 'SCALAR', 'x': 'class', 'y': 'Mean Average Precision'}]},
            'info': samples_scores}


class AntROCandAUCDet(AntMeasure):
    def __init__(self, task):
        super(AntROCandAUCDet, self).__init__(task, 'ROC_AUC')
        assert(task.task_type == 'OBJECT-DETECTION')

    def eva(self, data, label):
        category_num = len(self.task.class_label)
        detection_score = [[] for _ in range(category_num)]
        detection_label = [[] for _ in range(category_num)]

        if label is not None:
            data = zip(data, label)

        # assert(len(data) == len(label))
        #
        overlap_thre = float(getattr(self.task, 'roc_auc_overlap', 0.5))

        # 1.step positive sample is overlap > overlap_thre
        predict_box_total = 0
        for predict, gt in data:
            if predict is None:
                for missed_gt_bi in range(len(gt['bbox'])):
                    detection_label[gt['category_id'][missed_gt_bi]].append(1)
                    detection_score[gt['category_id'][missed_gt_bi]].append(-float("inf"))
                continue

            det_bbox = predict['det-bbox']
            det_score = predict['det-score']
            det_label = predict['det-label']

            # ground truth bbox and categories
            gt_bbox = np.array(gt['bbox'])
            gt_category = np.array(gt['category_id']).astype(dtype=np.int32)
            gt_bbox_num = gt_bbox.shape[0]

            # predict position, category, and score
            predict_box = np.array(det_bbox)
            predict_category = np.array(det_label).astype(dtype=np.int32)
            predict_score = np.array(det_score)
            predict_box_total += predict_score.shape[0]

            overlaps = bbox_overlaps(
                np.ascontiguousarray(predict_box, dtype=np.float),
                np.ascontiguousarray(gt_bbox, dtype=np.float))

            # distinguish positive and negative samples and scores (overlap > 0.5)
            gtm = np.ones((gt_bbox_num)) * (-1)
            for dind, d in enumerate(predict_box.tolist()):
                # information about best match so far (m=-1 -> unmatched)
                m = -1
                iou = 0.5
                for gind, g in enumerate(gt_bbox.tolist()):
                    if gt_category[gind] != predict_category[dind]:
                        continue

                    # if this gt already matched  continue
                    if gtm[gind] >= 0:
                        continue

                    # continue to next gt unless better match made
                    if overlaps[dind, gind] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = overlaps[dind, gind]
                    m = gind

                # if match made store id of match for both dt and gt
                if m > -1:
                    gtm[m] = dind

                    # success to match
                    detection_score[gt_category[m]].append(predict_score[dind])
                    detection_label[gt_category[m]].append(1)

            # process none matched det
            for dind, d in enumerate(predict_box.tolist()):
                if dind not in gtm:
                    detection_score[predict_category[dind]].append(predict_score[dind])
                    detection_label[predict_category[dind]].append(0)

            # process missed gt bbox
            missed_gt_bbox = [_ for _ in range(gt_bbox_num) if gtm[_] == -1]
            for missed_gt_bi in missed_gt_bbox:
                detection_label[gt_category[missed_gt_bi]].append(1)
                detection_score[gt_category[missed_gt_bi]].append(-float("inf"))

            #########################################################################

        # 2.step compute ROC curve and AUC
        category_roc_curves = []
        category_auc_scroes = []
        for predict, gt in zip(detection_label, detection_score):
            # skip 0
          
            roc_curve = roc(predict, gt)
            if roc_curve is None:
                roc_curve = np.array(())
            auc_score = auc(predict, gt)
            if auc_score is None:
                auc_score = 0
            category_roc_curves.append(roc_curve.tolist())
            category_auc_scroes.append(float(auc_score))

        return {'statistic': {'name': self.name,
                              'value': [{'name': 'ROC',
                                         'value': category_roc_curves,
                                         'type': 'CURVE',
                                         'x': 'FP',
                                         'y': 'TP'},
                                        {'name': 'AUC',
                                         'value': category_auc_scroes,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'AUC'}]},}


class AntPRDet(AntMeasure):
    def __init__(self, task):
        super(AntPRDet, self).__init__(task, 'PR')
        assert(task.task_type == 'OBJECT-DETECTION')

    def eva(self, data, label):
        category_num = len(self.task.class_label)
        detection_score = [[] for _ in range(category_num)]
        detection_label = [[] for _ in range(category_num)]

        if label is not None:
            data = zip(data, label)

        # assert(len(data) == len(label))

        #
        overlap_thre = float(getattr(self.task, 'pr_overlap', 0.5))

        # 1.step positive sample is overlap > overlap_thre
        predict_box_total = 0
        for predict, gt in data:
            if predict is None:
                for missed_gt_bi in range(len(gt['bbox'])):
                    detection_label[gt['category_id'][missed_gt_bi]].append(1)
                    detection_score[gt['category_id'][missed_gt_bi]].append(-float("inf"))
                continue

            det_bbox = predict['det-bbox']
            det_score = predict['det-score']
            det_label = predict['det-label']

            # ground truth bbox and categories
            gt_bbox = np.array(gt['bbox'])
            gt_category = np.array(gt['category_id']).astype(dtype=np.int32)
            gt_bbox_num = gt_bbox.shape[0]

            # predict position, category, and score
            predict_box = np.array(det_bbox)
            predict_category = np.array(det_label).astype(dtype=np.int32)
            predict_score = np.array(det_score)
            predict_box_total += predict_score.shape[0]

            overlaps = bbox_overlaps(
                np.ascontiguousarray(predict_box, dtype=np.float),
                np.ascontiguousarray(gt_bbox, dtype=np.float))

            # distinguish positive and negative samples and scores (overlap > 0.5)
            gtm = np.ones((gt_bbox_num)) * (-1)
            for dind, d in enumerate(predict_box.tolist()):
                # information about best match so far (m=-1 -> unmatched)
                m = -1
                iou = 0.5
                for gind, g in enumerate(gt_bbox.tolist()):
                    if gt_category[gind] != predict_category[dind]:
                        continue

                    # if this gt already matched  continue
                    if gtm[gind] >= 0:
                        continue

                    # continue to next gt unless better match made
                    if overlaps[dind, gind] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = overlaps[dind, gind]
                    m = gind

                # if match made store id of match for both dt and gt
                if m > -1:
                    gtm[m] = dind

                    # success to match
                    detection_score[gt_category[m]].append(predict_score[dind])
                    detection_label[gt_category[m]].append(1)

            # process none matched det
            for dind, d in enumerate(predict_box.tolist()):
                if dind not in gtm:
                    detection_score[predict_category[dind]].append(predict_score[dind])
                    detection_label[predict_category[dind]].append(0)

            # process missed gt bbox
            missed_gt_bbox = [_ for _ in range(gt_bbox_num) if gtm[_] == -1]
            for missed_gt_bi in missed_gt_bbox:
                detection_label[gt_category[missed_gt_bi]].append(1)
                detection_score[gt_category[missed_gt_bi]].append(-float("inf"))

            #########################################################################

        # 2.step compute Precicion Recall curve
        category_pr_curves = []
        for predict, gt in zip(detection_label, detection_score):
            # skip 0
          
            pr_curve = pr(predict, gt)
            if pr_curve is None:
                pr_curve = np.array(())
            category_pr_curves.append(pr_curve.tolist())

        return {'statistic': {'name': self.name,
                              'value': [{'name': 'Precision-Recall',
                                         'value': category_pr_curves,
                                         'type': 'CURVE',
                                         'x': 'recall',
                                         'y': 'precision'}]},}


class AntAPRFDet(AntMeasure):
    def __init__(self, task):
        super(AntAPRFDet, self).__init__(task, 'APRF')
        assert(task.task_type == 'OBJECT-DETECTION')

    def eva(self, data, label):
        category_num = len(self.task.class_label)
        detection_score = [[] for _ in range(category_num)]
        detection_label = [[] for _ in range(category_num)]

        if label is not None:
            data = zip(data, label)

        # assert(len(data) == len(label))

        #
        overlap_thre = float(getattr(self.task, 'aprf_overlap', 0.5))

        # 1.step positive sample is overlap > overlap_thre
        predict_box_total = 0
        for predict, gt in data:
            if predict is None:
                for missed_gt_bi in range(len(gt['bbox'])):
                    detection_label[gt['category_id'][missed_gt_bi]].append(1)
                    detection_score[gt['category_id'][missed_gt_bi]].append(-float("inf"))
                continue

            det_bbox = predict['det-bbox']
            det_score = predict['det-score']
            det_label = predict['det-label']

            # ground truth bbox and categories
            gt_bbox = np.array(gt['bbox'])
            gt_category = np.array(gt['category_id']).astype(dtype=np.int32)
            gt_bbox_num = gt_bbox.shape[0]

            # predict position, category, and score
            predict_box = np.array(det_bbox)
            predict_category = np.array(det_label).astype(dtype=np.int32)
            predict_score = np.array(det_score)
            predict_box_total += predict_score.shape[0]

            overlaps = bbox_overlaps(
                np.ascontiguousarray(predict_box, dtype=np.float),
                np.ascontiguousarray(gt_bbox, dtype=np.float))

            # distinguish positive and negative samples and scores (overlap > 0.5)
            gtm = np.ones((gt_bbox_num)) * (-1)
            for dind, d in enumerate(predict_box.tolist()):
                # information about best match so far (m=-1 -> unmatched)
                m = -1
                iou = 0.5
                for gind, g in enumerate(gt_bbox.tolist()):
                    if gt_category[gind] != predict_category[dind]:
                        continue

                    # if this gt already matched  continue
                    if gtm[gind] >= 0:
                        continue

                    # continue to next gt unless better match made
                    if overlaps[dind, gind] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = overlaps[dind, gind]
                    m = gind

                # if match made store id of match for both dt and gt
                if m > -1:
                    gtm[m] = dind

                    # success to match
                    detection_score[gt_category[m]].append(predict_score[dind])
                    detection_label[gt_category[m]].append(1)

            # process none matched det
            for dind, d in enumerate(predict_box.tolist()):
                if dind not in gtm:
                    detection_score[predict_category[dind]].append(predict_score[dind])
                    detection_label[predict_category[dind]].append(0)

            # process missed gt bbox
            missed_gt_bbox = [_ for _ in range(gt_bbox_num) if gtm[_] == -1]
            for missed_gt_bi in missed_gt_bbox:
                detection_label[gt_category[missed_gt_bi]].append(1)
                detection_score[gt_category[missed_gt_bi]].append(-float("inf"))

            #########################################################################

        # 2.step compute accuracy, precision, recall, and F1
        category_accuracy = []
        category_precision = []
        category_recall = []
        category_F1 = []
        for category_detection_label, category_detection_score in zip(detection_label, detection_score):
            # skip 0
          
            predicated_label = [1 if s > -float("inf") else 0 for s in category_detection_score]

            res = binary_c_stats2(actual=category_detection_label, predicated=predicated_label)
            accuracy, precision, recall, F1 = res if res is not None else 0, 0, 0, 0

            category_accuracy.append(accuracy)
            category_precision.append(precision)
            category_recall.append(recall)
            category_F1.append(F1)

        return {'statistic': {'name': self.name,
                              'value': [{'name': 'Accuracy',
                                         'value': category_accuracy,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'accuracy'},
                                        {'name': 'Precision',
                                         'value': category_precision,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'precision'},
                                        {'name': 'Recall',
                                         'value': category_recall,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'recall'},
                                        {'name': 'F1',
                                         'value': category_F1,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'F1'}]}}


class AntTFTFDet(AntMeasure):
    def __init__(self, task):
        super(AntTFTFDet, self).__init__(task, 'TFTF')
        assert(task.task_type == 'OBJECT-DETECTION')

    def eva(self, data, label):
        category_num = len(self.task.class_label)
        detection_score = [[] for _ in range(category_num)]
        detection_label = [[] for _ in range(category_num)]

        if label is not None:
            data = zip(data, label)
        # assert(len(data) == len(label))

        #
        overlap_thre = float(getattr(self.task,'APRF_overlap',0.5))

        # 1.step positive sample is overlap > overlap_thre
        predict_box_total = 0
        for predict, gt in data:
            if predict is None:
                for missed_gt_bi in range(len(gt['bbox'])):
                    detection_label[gt['category_id'][missed_gt_bi]].append(1)
                    detection_score[gt['category_id'][missed_gt_bi]].append(-float("inf"))
                continue

            det_bbox = predict['det-bbox']
            det_score = predict['det-score'].flatten()
            det_label = predict['det-label'].flatten()

            # ground truth bbox and categories
            gt_bbox = np.array(gt['bbox'])
            gt_category = np.array(gt['category_id']).astype(dtype=np.int32)
            gt_bbox_num = gt_bbox.shape[0]

            # predict position, category, and score
            predict_box = np.array(det_bbox)
            predict_category = np.array(det_label).astype(dtype=np.int32)
            predict_score = np.array(det_score)
            predict_box_total += predict_score.shape[0]

            overlaps = bbox_overlaps(
                np.ascontiguousarray(predict_box, dtype=np.float),
                np.ascontiguousarray(gt_bbox, dtype=np.float))

            # distinguish positive and negative samples and scores (overlap > 0.5)
            gtm = np.ones((gt_bbox_num)) * (-1)
            for dind, d in enumerate(predict_box.tolist()):
                # information about best match so far (m=-1 -> unmatched)
                m = -1
                iou = 0.5
                for gind, g in enumerate(gt_bbox.tolist()):
                    if gt_category[gind] != predict_category[dind]:
                        continue

                    # if this gt already matched  continue
                    if gtm[gind] >= 0:
                        continue

                    # continue to next gt unless better match made
                    if overlaps[dind, gind] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = overlaps[dind, gind]
                    m = gind

                # if match made store id of match for both dt and gt
                if m > -1:
                    gtm[m] = dind

                    # success to match
                    detection_score[gt_category[m]].append(predict_score[dind])
                    detection_label[gt_category[m]].append(1)

            # process none matched det
            for dind, d in enumerate(predict_box.tolist()):
                if dind not in gtm:
                    detection_score[predict_category[dind]].append(predict_score[dind])
                    detection_label[predict_category[dind]].append(0)

            # process missed gt bbox
            missed_gt_bbox = [_ for _ in range(gt_bbox_num) if gtm[_] == -1]
            for missed_gt_bi in missed_gt_bbox:
                detection_label[gt_category[missed_gt_bi]].append(1)
                detection_score[gt_category[missed_gt_bi]].append(-float("inf"))

            #########################################################################

        # 2.step compute true positive(TP), false negative(FN), true negative(TN), false positive(FP)
        category_TP = []
        category_FN = []
        category_TN = []
        category_FP = []
        for category_detection_label,category_detection_score in zip(detection_label,detection_score):
            # skip 0
            predicated_label = [1 if s > -float("inf") else 0 for s in category_detection_score]

            res = binary_c_stats(actual=category_detection_label, predicated=predicated_label)
            TP,FN,TN,FP = res if res is not None else 0, 0, 0, 0

            category_TP.append(TP)
            category_FN.append(FN)
            category_TN.append(TN)
            category_FP.append(FP)

        return {'statistic': {'name': self.name,
                              'value': [{'name': 'true-positive',
                                         'value': category_TP,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'TP'},
                                        {'name': 'false-negative',
                                         'value': category_FN,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'FN'},
                                        {'name': 'true-negative',
                                         'value': category_TN,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'TN'},
                                        {'name': 'false-positive',
                                         'value': category_FP,
                                         'type': 'SCALAR',
                                         'x': 'class',
                                         'y': 'FP'}]},}


class AntCOCODet(AntMeasure):
    def __init__(self, task):
        super(AntCOCODet, self).__init__(task, 'COCO')
        assert(task.task_type == 'OBJECT-DETECTION')

        # np.arange causes trouble.  the data point on arange is slightly larger than the true value
        self.iouThrs = np.linspace(.5, 0.95, np.round((0.95 - .5) / .05) + 1, endpoint=True)
        self.recThrs = np.linspace(.0, 1.00, np.round((1.00 - .0) / .01) + 1, endpoint=True)
        self.maxDets = [1, 10, 100]
        self.areaRng = [[0 ** 2, 1e5 ** 2], [0 ** 2, 32 ** 2], [32 ** 2, 96 ** 2], [96 ** 2, 1e5 ** 2]]
        self.areaRngLbl = ['all', 'small', 'medium', 'large']
        self.useCats = 1
        self.dts = []
        self.gts = []
        self.imgIds = []
        self.catIds = []

    def _parse_result(self, data, label):
        if label is not None:
            data = zip(data, label)

        # assert(len(data) == len(label))
        gt_id = 1
        det_id = 1
        for index, (a, b) in enumerate(data):
            image_id = -1
            for gt_box_i, gt_box in enumerate(b['bbox']):
                gt = {}
                x0, y0, x1, y1 = gt_box
                gt['bbox'] = np.array([x0, y0, x1, y1])
                gt['area'] = b['area'][gt_box_i]
                gt['image_id'] = b['id']
                gt['category_id'] = b['category_id'][gt_box_i]
                gt['iscrowd'] = b['iscrowd'][gt_box_i] if 'iscrowd' in b else 0
                gt['id'] = gt_id
                gt_id += 1
                image_id = b['id']
                self.imgIds.append(image_id)
                self.catIds.append(b['category_id'][gt_box_i])
                self.gts.append(gt)

            if a is None:
                dt = {}
                dt['bbox'] = None
                dt['area'] = 0
                dt['image_id'] = -1
                dt['category_id'] = -1
                dt['score'] = -1
                dt['id'] = -1
                self.dts.append(dt)
                continue

            # a dict {'det-bbox':, 'det-score':, 'det-label':,}
            # b annotation
            det_bbox = a['det-bbox']
            det_score = a['det-score'].flatten()
            det_label = a['det-label'].flatten()

            for box_i, box in enumerate(det_bbox):
                dt = {}
                x0, y0, x1, y1 = box
                score = det_score[box_i]
                label = int(det_label[box_i])

                dt['bbox'] = np.array([x0, y0, x1, y1])
                dt['area'] = (x1 - x0) * (y1 - y0)
                dt['image_id'] = image_id
                dt['category_id'] = int(label)
                dt['score'] = score
                dt['id'] = det_id
                det_id += 1

                self.dts.append(dt)

    def _prepare(self):
        '''
        Prepare ._gts and ._dts for evaluation based on params
        :return: None
        '''
        # set ignore flag
        for gt in self.gts:
            gt['ignore'] = gt['ignore'] if 'ignore' in gt else 0
            gt['ignore'] = 'iscrowd' in gt and gt['iscrowd']

        self._gts = defaultdict(list)       # gt for evaluation
        self._dts = defaultdict(list)       # dt for evaluation
        for gt in self.gts:
            self._gts[gt['image_id'], gt['category_id']].append(gt)
        for dt in self.dts:
            self._dts[dt['image_id'], dt['category_id']].append(dt)
        self.evalImgs = defaultdict(list)   # per-image per-category evaluation results
        self.eval     = {}                  # accumulated evaluation results

    def _evaluate(self):
        '''
        Run per image evaluation on given images and store results (a list of dict) in self.evalImgs
        :return: None
        '''
        #p = self.params

        self.imgIds = list(np.unique(self.imgIds))
        self.catIds = list(np.unique(self.catIds))
        self.maxDets = sorted(self.maxDets)

        self._prepare()
        # loop through images, area range, max detection number
        catIds = self.catIds
        computeIoU = self._computeIoU

        self.ious = {(imgId, catId): computeIoU(imgId, catId) \
                        for imgId in self.imgIds
                        for catId in catIds}

        maxDet = self.maxDets[-1]
        self.evalImgs = [self._evaluateImg(imgId, catId, areaRng, maxDet)
                 for catId in catIds
                 for areaRng in self.areaRng
                 for imgId in self.imgIds
             ]

    def _computeIoU(self, imgId, catId):
        # p = self.params
        gt = self._gts[imgId, catId]
        dt = self._dts[imgId, catId]

        if len(gt) == 0 and len(dt) == 0:
            return []
        inds = np.argsort([-d['score'] for d in dt], kind='mergesort')
        dt = [dt[i] for i in inds]
        if len(dt) > self.maxDets[-1]:
            dt = dt[0:self.maxDets[-1]]

        g = [g['bbox'] for g in gt]
        d = [d['bbox'] for d in dt]

        # compute iou between each dt and gt region
        iscrowd = [int(o['iscrowd']) for o in gt]
        ious = _mask.iou(d, g, iscrowd)
        return ious

    def _evaluateImg(self, imgId, catId, aRng, maxDet):
        '''
        perform evaluation for single category and image
        :return: dict (single image results)
        '''
        gt = self._gts[imgId, catId]
        dt = self._dts[imgId, catId]

        if len(gt) == 0 and len(dt) ==0:
            return None

        for g in gt:
            if g['ignore'] or (g['area'] < aRng[0] or g['area'] > aRng[1]):
                g['_ignore'] = 1
            else:
                g['_ignore'] = 0

        # sort dt highest score first, sort gt ignore last
        gtind = np.argsort([g['_ignore'] for g in gt], kind='mergesort')
        gt = [gt[i] for i in gtind]
        dtind = np.argsort([-d['score'] for d in dt], kind='mergesort')
        dt = [dt[i] for i in dtind[0:maxDet]]
        iscrowd = [int(o['iscrowd']) for o in gt]
        # load computed ious
        ious = self.ious[imgId, catId][:, gtind] if len(self.ious[imgId, catId]) > 0 else self.ious[imgId, catId]

        T = len(self.iouThrs)
        G = len(gt)
        D = len(dt)
        gtm  = np.zeros((T,G))
        dtm  = np.zeros((T,D))
        gtIg = np.array([g['_ignore'] for g in gt])
        dtIg = np.zeros((T,D))
        if not len(ious)==0:
            for tind, t in enumerate(self.iouThrs):
                for dind, d in enumerate(dt):
                    # information about best match so far (m=-1 -> unmatched)
                    iou = min([t, 1-1e-10])
                    m   = -1
                    for gind, g in enumerate(gt):
                        # if this gt already matched, and not a crowd, continue
                        if gtm[tind, gind] > 0 and not iscrowd[gind]:
                            continue
                        # if dt matched to reg gt, and on ignore gt, stop
                        if m > -1 and gtIg[m] == 0 and gtIg[gind] == 1:
                            break
                        # continue to next gt unless better match made
                        if ious[dind, gind] < iou:
                            continue
                        # if match successful and best so far, store appropriately
                        iou = ious[dind, gind]
                        m = gind
                    # if match made store id of match for both dt and gt
                    if m == -1:
                        continue
                    dtIg[tind, dind] = gtIg[m]
                    dtm[tind, dind]  = gt[m]['id']
                    gtm[tind, m]     = d['id']
        # set unmatched detections outside of area range to ignore
        a = np.array([d['area']<aRng[0] or d['area']>aRng[1] for d in dt]).reshape((1, len(dt)))
        dtIg = np.logical_or(dtIg, np.logical_and(dtm==0, np.repeat(a,T,0)))
        # store results for given image and category
        return {
                'image_id':     imgId,
                'category_id':  catId,
                'aRng':         aRng,
                'maxDet':       maxDet,
                'dtIds':        [d['id'] for d in dt],
                'gtIds':        [g['id'] for g in gt],
                'dtMatches':    dtm,
                'gtMatches':    gtm,
                'dtScores':     [d['score'] for d in dt],
                'gtIgnore':     gtIg,
                'dtIgnore':     dtIg,
            }

    def _accumulate(self):
        '''
        Accumulate per image evaluation results and store the result in self.eval
        :param p: input params for evaluation
        :return: None
        '''
        if not self.evalImgs:
            print('Please run evaluate() first')
        # allows input customized parameters
        # if p is None:
        #     p = self.params

        T           = len(self.iouThrs)
        R           = len(self.recThrs)
        K           = len(self.catIds)
        A           = len(self.areaRng)
        M           = len(self.maxDets)
        precision   = -np.ones((T,R,K,A,M)) # -1 for the precision of absent categories
        recall      = -np.ones((T,K,A,M))

        # create dictionary for future indexing
        catIds = self.catIds
        setK = set(catIds)
        setA = set(map(tuple, self.areaRng))
        setM = set(self.maxDets)
        setI = set(self.imgIds)
        # get inds to evaluate
        k_list = [n for n, k in enumerate(self.catIds)  if k in setK]
        m_list = [m for n, m in enumerate(self.maxDets) if m in setM]
        a_list = [n for n, a in enumerate(map(lambda x: tuple(x), self.areaRng)) if a in setA]
        i_list = [n for n, i in enumerate(self.imgIds)  if i in setI]
        I0 = len(self.imgIds)
        A0 = len(self.areaRng)
        # retrieve E at each category, area range, and max number of detections
        for k, k0 in enumerate(k_list):
            Nk = k0*A0*I0
            for a, a0 in enumerate(a_list):
                Na = a0*I0
                for m, maxDet in enumerate(m_list):
                    E = [self.evalImgs[Nk + Na + i] for i in i_list]
                    E = [e for e in E if not e is None]
                    if len(E) == 0:
                        continue
                    dtScores = np.concatenate([e['dtScores'][0:maxDet] for e in E])

                    # different sorting method generates slightly different results.
                    # mergesort is used to be consistent as Matlab implementation.
                    inds = np.argsort(-dtScores, kind='mergesort')

                    dtm  = np.concatenate([e['dtMatches'][:,0:maxDet] for e in E], axis=1)[:,inds]
                    dtIg = np.concatenate([e['dtIgnore'][:,0:maxDet]  for e in E], axis=1)[:,inds]
                    gtIg = np.concatenate([e['gtIgnore'] for e in E])
                    npig = np.count_nonzero(gtIg==0 )
                    if npig == 0:
                        continue
                    tps = np.logical_and(               dtm,  np.logical_not(dtIg) )
                    fps = np.logical_and(np.logical_not(dtm), np.logical_not(dtIg) )

                    tp_sum = np.cumsum(tps, axis=1).astype(dtype=np.float)
                    fp_sum = np.cumsum(fps, axis=1).astype(dtype=np.float)
                    for t, (tp, fp) in enumerate(zip(tp_sum, fp_sum)):
                        tp = np.array(tp)
                        fp = np.array(fp)
                        nd = len(tp)
                        rc = tp / npig
                        pr = tp / (fp+tp+np.spacing(1))
                        q  = np.zeros((R,))

                        if nd:
                            recall[t,k,a,m] = rc[-1]
                        else:
                            recall[t,k,a,m] = 0

                        # numpy is slow without cython optimization for accessing elements
                        # use python array gets significant speed improvement
                        pr = pr.tolist(); q = q.tolist()

                        for i in range(nd-1, 0, -1):
                            if pr[i] > pr[i-1]:
                                pr[i-1] = pr[i]

                        inds = np.searchsorted(rc, self.recThrs, side='left')
                        try:
                            for ri, pi in enumerate(inds):
                                if pi == len(pr):
                                    q[ri] = pr[pi - 1]
                                else:
                                    q[ri] = pr[pi]
                        except:
                            pass

                        precision[t,:,k,a,m] = np.array(q)
        self.eval = {
            'counts': [T, R, K, A, M],
            'precision': precision,
            'recall':   recall,
        }

    def eva(self, data, label):
        self._parse_result(data, label)
        self._evaluate()
        self._accumulate()

        def _summarize(ap=1, iouThr=None, areaRng='all', maxDets=100 ):
            iStr = ' {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}'
            titleStr = 'Average Precision' if ap == 1 else 'Average Recall'
            typeStr = '(AP)' if ap==1 else '(AR)'
            iouStr = '{:0.2f}:{:0.2f}'.format(self.iouThrs[0], self.iouThrs[-1]) \
                if iouThr is None else '{:0.2f}'.format(iouThr)

            aind = [i for i, aRng in enumerate(self.areaRngLbl) if aRng == areaRng]
            mind = [i for i, mDet in enumerate(self.maxDets) if mDet == maxDets]
            if ap == 1:
                # dimension of precision: [TxRxKxAxM]
                s = self.eval['precision']
                # IoU
                if iouThr is not None:
                    t = np.where(iouThr == self.iouThrs)[0]
                    s = s[t]
                s = s[:,:,:,aind,mind]
            else:
                # dimension of recall: [TxKxAxM]
                s = self.eval['recall']
                if iouThr is not None:
                    t = np.where(iouThr == self.iouThrs)[0]
                    s = s[t]
                s = s[:,:,aind,mind]
            if len(s[s>-1])==0:
                mean_s = -1
            else:
                mean_s = np.mean(s[s>-1])

            return mean_s, iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s)

        # all
        coco_status, coco_status_str = _summarize(1)
        # iou-thre = .5, maxdets = 100
        coco_iou05_maxdets100, coco_iou05_maxdets100_str = \
            _summarize(1, iouThr=.5, maxDets=self.maxDets[2])
        # iou-thre = .75, maxdets = 100
        coco_iou075_maxdets100, coco_iou075_maxdets100_str = \
            _summarize(1, iouThr=.75, maxDets=self.maxDets[2])
        # area-small, maxdets = 100
        coco_small_maxdets100, coco_small_maxdets100_str = \
            _summarize(1, areaRng='small', maxDets=self.maxDets[2])
        # area-medium, maxdets = 100
        coco_medium_maxdets100, coco_medium_maxdets100_str = \
            _summarize(1, areaRng='medium', maxDets=self.maxDets[2])
        # area-large, maxdets = 100
        coco_large_maxdets100, coco_large_maxdets100_str = \
            _summarize(1, areaRng='large', maxDets=self.maxDets[2])
        # maxdets = 1
        coco_maxdets1, coco_maxdets1_str = \
            _summarize(0, maxDets=self.maxDets[0])
        # maxdets = 10
        coco_maxdets10, coco_maxdets10_str = \
            _summarize(0, maxDets=self.maxDets[1])
        # maxdets = 100
        coco_maxdets100, coco_maxdets100_str = \
            _summarize(0, maxDets=self.maxDets[2])
        # area-small, maxdets = 100
        coco_small_maxdets100_nap, coco_small_maxdets100_nap_str = \
            _summarize(0, areaRng='small', maxDets=self.maxDets[2])
        # area-medium, maxdets = 100
        coco_medium_maxdets100_nap, coco_medium_maxdets100_nap_str = \
            _summarize(0, areaRng='medium', maxDets=self.maxDets[2])
        # area-large, maxdets = 100
        coco_large_maxdets100_nap, coco_large_maxdets100_nap_str = \
            _summarize(0, areaRng='large', maxDets=self.maxDets[2])

        return {'statistic':{'name':self.name,
                             'value':[{'name': self.name,
                                       'value': coco_status,
                                       'type': 'SCALAR',
                                       'x': coco_status_str},
                                      {'name': 'iou-0.5-max-det-100-AP',
                                       'value': coco_iou05_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_iou05_maxdets100_str},
                                      {'name': 'iou-0.75-max-det-100-AP',
                                       'value': coco_iou075_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_iou075_maxdets100_str},
                                      {'name': 'area-small-AP',
                                       'value': coco_small_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_small_maxdets100_str},
                                      {'name': 'area-medium-AP',
                                       'value': coco_medium_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_medium_maxdets100_str},
                                      {'name': 'area-large-AP',
                                       'value': coco_large_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_large_maxdets100_str},
                                      {'name': 'max-det-1',
                                       'value': coco_maxdets1,
                                       'type': 'SCALAR',
                                       'x': coco_maxdets1_str},
                                      {'name': 'max-det-10',
                                       'value': coco_maxdets10,
                                       'type': 'SCALAR',
                                       'x': coco_maxdets10_str},
                                      {'name': 'max-det-100',
                                       'value': coco_maxdets100,
                                       'type': 'SCALAR',
                                       'x': coco_maxdets100_str},
                                      {'name': 'area-small',
                                       'value': coco_small_maxdets100_nap,
                                       'type': 'SCALAR',
                                       'x': coco_small_maxdets100_nap_str},
                                      {'name': 'area-medium',
                                       'value': coco_medium_maxdets100_nap,
                                       'type': 'SCALAR',
                                       'x': coco_medium_maxdets100_nap_str},
                                      {'name': 'area-large',
                                       'value': coco_large_maxdets100_nap,
                                       'type': 'SCALAR',
                                       'x': coco_large_maxdets100_nap_str}]},}


# -*- coding: UTF-8 -*-
# @Time    : 2022/8/21 21:46
# @File    : label_server.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import sys
import copy

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import sys
import uuid
import json
import time
import math
import numpy as np
from io import BytesIO
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor
import threading
import base64
import imagesize
from antgo.utils import colormap


MB = 1024 * 1024
GB = 1024 * MB
MAX_STREAMED_SIZE = 1*GB

class LiveApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class LabelTaskHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    task_type = self.db['task_metas']['task_type']
    task_name = self.db['task_metas']['task_name']
    label_type = self.db['task_metas']['label_type']

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'task_type': task_type,
      'task_name': task_name,
      'label_type': label_type,
      'operator': {
        'full_name': user['full_name'],
        'short_name': user['short_name'],
      },
      'label_metas': self.db['label_metas']
    })


class LabelStateHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    if 'running' not in self.db:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    info = {
      'project_type': 'LABEL',
      'project_state': {
        'state': self.db['running']['state'],
        'stage': self.db['running']['stage'],
        'round': self.db['running']['round'],
        'waiting_time_to_next_round': self.db['running']['waiting_time_to_next_round']
      }
    }

    if self.db['running']['state'] == 'running' and self.db['running']['stage'] == 'labeling':
      sample_num_completed = 0
      for sample in self.db['samples']:
        if sample['completed_time'] != 0:
          sample_num_completed += 1
      info['project_state'].update({
        'sample_num_completed': sample_num_completed,
        'sample_num': len(self.db['samples'])
      })

    # state: 'running'/'abnormal'/'overtime'/'pending'
    # stage: 'waiting'/'training'/'labeling'/'finish'
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=info)

  @gen.coroutine
  def post(self):
    running_state = self.get_argument('running_state', None)
    running_stage = self.get_argument('running_stage', None)
    running_round = self.get_argument("running_round", '-1')
    if running_state is None or running_stage is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    assert(running_state in ['running', 'abnormal', 'overtime', 'pending'])
    assert(running_stage in ['waiting', 'labeling', 'finish'])
    # running_stage的finish状态需要由前端触发，标记本轮标注已经结束

    if running_stage == 'labeling':
      running_round = int(running_round)
      if running_round == -1:
        self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
        return 

      self.db['running']['state'] = running_state
      self.db['running']['stage'] = running_stage
      self.db['running']['round'] = (int)(running_round)
      self.response(RESPONSE_STATUS_CODE.SUCCESS)
      return

    # 强制完成标记 （如果是true，则即使是未完成所有样本标注，也强制设置完成标记）
    is_force = self.get_argument('force', False)
    if running_stage == 'finish':
      # 检查是否每个数据已经完成审核或标注
      sample_num = len(self.db['samples'])
      completed_num = 0
      for sample in self.db['samples']:
        if sample['state'] == 'completed':
          completed_num += 1
        
      if sample_num != completed_num and not is_force:
        self.response(
          RESPONSE_STATUS_CODE.SUCCESS, 
          content={
            'finish': False,
            'completed_num': completed_num,
            'sample_num': sample_num
          })
        return 
      else:
        self.response(
          RESPONSE_STATUS_CODE.SUCCESS, 
          content={
            'finish': True
          })
  
    if running_stage == 'waiting':
      # 切换运行状态为等待, 清空标注样本
      self.db['samples'] = []

    self.db['running']['state'] = running_state
    self.db['running']['stage'] = running_stage
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class LabelSamplesHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 获得所有等待标注数据
    page_index = self.get_argument('page_index', None)
    page_size = self.get_argument('page_size', None)
    index_offset = self.get_argument('index_offset', None)

    samples = []
    if page_size is None or page_index is None:
      samples = self.db['samples']
    elif index_offset is not None:
      index_offset = (int)(index_offset)
      page_size = (int)(page_size)
      start_page_index = (int)(page_index)
      end_page_index = index_offset // page_size + 1

      start_index = start_page_index * page_size
      end_index = end_page_index * page_size

      samples = self.db['samples'][start_index:end_index]
      page_index = end_page_index
    else:
      page_index = (int)(page_index)
      page_size = (int)(page_size)
      start_index = page_index * page_size
      end_index = (page_index + 1) * page_size

      samples = self.db['samples'][start_index:end_index]
      page_index = page_index + 1

    # copy
    samples = copy.deepcopy(samples)
    for sample in samples:
      # 转换时间到可显示
      for key in sample.keys():
        if key.endswith('time'):
          sample[key] = self.timestamp_2_str(sample[key])

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'page_samples': samples,
      'page_index': page_index,
      'columns': self.db['sample_metas']['filters'],
      'total_sample_num': len(self.db['samples'])
    })


class LabelFreshSampleHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 样本例子
    # {
    # 'image_file': '../assets/1.jpeg',
    # 'height': 800,
    # 'width': 1200,
    # 'sample_id': 2,
    # 'completed_time': 0,
    # 'created_time': 0,
    # 'update_time': 0,
    # 'state': 'waiting',
    # 'assigner': '',
    # 'operators': [],
    # 'label_info': [],
    # }
    #
    # 添加样本记录
    samples = self.get_argument('samples', None)
    if samples is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return
    samples = json.loads(samples)

    offset = len(self.db['samples'])
    for sample_i, sample in enumerate(samples):
      image_file = sample['image_file']
      # image_path = os.path.join(self.settings.get('static_path'), image_file)
      # width, height = imagesize.get(image_path)
      # sample['width'] = width
      # sample['height'] = height
      sample['sample_id'] = sample_i + offset
      sample['completed_time'] = 0
      sample['created_time'] = 0
      sample['update_time'] = 0
      sample['state'] = 'waiting'
      sample['assigner'] = ''
      sample['operators'] = []
      if 'label_info' not in sample:
        sample['label_info'] = []

    # 更新到数据库中
    self.db['samples'].extend(samples)
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class LabelUpdateSampleHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 样本标注结果
    update_sample = self.get_argument('sample', None)
    if update_sample is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # # 当前操作者
    # user_name = self.get_argument('user_name', 'default')
    # if user_name is None:
    #   self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
    #   return

    update_sample = json.loads(update_sample)
    sample_id = update_sample['sample_id']
    if sample_id < 0 or sample_id >= len(self.db['samples']):
      # 异常
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if self.db['samples'][sample_id]['image_file'] != update_sample['image_file']:
      # 异常
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if user['labeling_sample'] == -1 or user['start_time'] == -1:
      # 异常
      self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
      return

    if self.db['samples'][sample_id]['assigner'] != user['full_name']:
      # 说明，此样本已经重新分配给了其他人，当前操作者不可修改
      self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
      return

    now_time = time.time()
    if now_time - user['start_time'] > 10*60:
      # 当前用户标注超时，不能更新结果
      user['labeling_sample'] = -1
      user['start_time'] = -1
      self.response(RESPONSE_STATUS_CODE.EXECUTE_FORBIDDEN)
      return

    #
    label_info = update_sample['label_info']
    self.db['samples'][sample_id]['label_info'] = label_info  # 标注信息记录(仅保留最新一次的标注更新)
    self.db['samples'][sample_id]['state'] = 'completed'      # 设置当前样本处在完成状态（可能当前样本已经处在完成状态）

    # 如果当前样本首次提交，则更新完成时间
    # 其余情况仅更新更新时间
    self.db['samples'][sample_id]['update_time'] = now_time
    if self.db['samples'][sample_id]['completed_time'] == 0:
      self.db['samples'][sample_id]['completed_time'] = now_time

    # 当更新后，每个样本需要记录操作者信息（名字，更新时间）
    pi = -1
    for i, p in enumerate(self.db['samples'][sample_id]['operators']):
      if p['full_name'] == user['full_name']:
        pi = i
    if pi == -1:
      # 为此样本，创建新的操作者
      self.db['samples'][sample_id]['operators'].append({
        'full_name': user['full_name'],
        'short_name': user['short_name'],
        'update_time': now_time,
        'create_time': now_time,
        'label_info': label_info
      })
    else:
      self.db['samples'][sample_id]['operators'][pi]['update_time'] = now_time
      self.db['samples'][sample_id]['operators'][pi]['label_info'] = label_info

    # finish
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'sample_id': sample_id,
      'sample_state': self.db['samples'][sample_id]['state'],
      'completed_time': self.timestamp_2_str(self.db['samples'][sample_id]['completed_time']),
      'update_time': self.timestamp_2_str(self.db['samples'][sample_id]['update_time']),
    })


class LabelGetSampleHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 调用此接口时，返回的样本是有可能已经存在标注结果
    sample_i = self.get_argument('sample_i', None)
    if sample_i is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    sample_i = (int)(sample_i)
    if sample_i < 0 or sample_i >= len(self.db['samples']):
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    now_time = time.time()
    if user['labeling_sample'] != -1:
      # 当前用户存在正在标注的样本
      now_labeling_sample = user['labeling_sample']
      if self.db['samples'][now_labeling_sample]['assigner'] == user['full_name'] and \
          self.db['samples'][now_labeling_sample]['state'] != 'completed':
        # 如果之前的那个样本还没有处在完成状态(包括标注状态)，则重新切换成等待状态
        self.db['samples'][now_labeling_sample]['state'] = 'waiting'
        self.db['samples'][now_labeling_sample]['created_time'] = 0
        self.db['samples'][now_labeling_sample]['assigner'] = ''

      # 重置
      user['labeling_sample'] = -1
      user['start_time'] = -1

    # 更改当前操作者的状态
    user['labeling_sample'] = sample_i
    user['start_time'] = now_time

    # 更改当前样本的状态
    if self.db['samples'][sample_i]['state'] != 'waiting':
      # 进入此处，说明处在标注状态或完成状态
      if self.db['samples'][sample_i]['state'] != 'completed':
        if now_time - self.db['samples'][sample_i]['created_time'] > 10 * 60:
          # 当前样本处在超时状态 (十分钟)
          # 将当前样本状态切换为等待状态
          self.db['samples'][sample_i]['state'] = 'waiting'     # 重置状态
          self.db['samples'][sample_i]['assigner'] = ''         # 清空已经分配的操作者
          self.db['samples'][sample_i]['created_time'] = 0      # 清空创建时间

    # 仅有当前样本处在未锁定状态才允许进行重新分配标注人员
    if self.db['samples'][sample_i]['state'] != 'labeling':
      self.db['samples'][sample_i]['assigner'] = user['full_name']

    if self.db['samples'][sample_i]['state'] == 'waiting':
      # 仅当当前样本状态处在等待状态时，需要改变当前状态为标注状态，并设置创建时间
      # 否则保持状态不变
      self.db['samples'][sample_i]['state'] = 'labeling'
      self.db['samples'][sample_i]['created_time'] = now_time

    sample = copy.deepcopy(self.db['samples'][sample_i])
    # 转换时间到可显示
    for key in sample.keys():
      if key.endswith('time'):
        sample[key] = self.timestamp_2_str(sample[key])

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'sample': sample
    })


class LabelNextSampleHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # TODO, 需要根据当前标注样本索引，继续向下寻找
    # update_sample = self.get_argument('sample', None)
  
    # 发现下一个还没有标注的样本(没有被分配的样本)
    waiting_label_sample_i = -1
    # state: labeling/completed/waiting
    # state: 标注状态/完成状态/等待状态/超时状态
    for sample_i in range(len(self.db['samples'])):
      if self.db['samples'][sample_i]['state'] == 'waiting':
        waiting_label_sample_i = sample_i
        break

      if self.db['samples'][sample_i]['state'] != 'completed':
        if time.time() - self.db['samples'][sample_i]['created_time'] > 10 * 60:
          # 当前样本处在超时状态 (十分钟)
          # 将当前样本状态切换为等待状态
          self.db['samples'][sample_i]['state'] = 'waiting'     # 重置状态
          self.db['samples'][sample_i]['assigner'] = ''         # 清空已经分配的操作者
          self.db['samples'][sample_i]['created_time'] = 0      # 清空创建时间
          waiting_label_sample_i = sample_i
          break

    if waiting_label_sample_i == -1:
      # 没有发现需要标注的样本
      # 说明当前轮标注结束，修改系统状态

      # self.db['running']['state'] = 'running'
      # self.db['running']['stage'] = 'finish'
      # self.db['running']['round'] += self.db['running']['round']
      # self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      #   'state': self.db['running']['state'],
      #   'stage': self.db['running']['stage'],
      #   'round': self.db['running']['round']
      # })
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if waiting_label_sample_i < 0 or waiting_label_sample_i >= len(self.db['samples']):
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    # 绑定当前用户的正在进行的标注样本
    if user['labeling_sample'] != -1:
      # 当前用户存在正在标注的样本
      now_labeling_sample = user['labeling_sample']
      if self.db['samples'][now_labeling_sample]['assigner'] == user['full_name'] and \
          self.db['samples'][now_labeling_sample]['state'] != 'completed':
        # 如果之前的那个样本还没有处在完成状态(包括标注状态,或等待状态)，则重新切换成等待状态
        self.db['samples'][now_labeling_sample]['state'] = 'waiting'
        self.db['samples'][now_labeling_sample]['created_time'] = 0
        self.db['samples'][now_labeling_sample]['assigner'] = ''

      # 重置
      user['labeling_sample'] = -1
      user['start_time'] = -1

    # 挑选出来的样本必须处在等待状态
    assert(self.db['samples'][waiting_label_sample_i]['state'] == 'waiting')

    now_time = time.time()
    # 更改当前操作者的状态
    user['labeling_sample'] = waiting_label_sample_i
    user['start_time'] = now_time

    # 更改当前样本的状态
    self.db['samples'][waiting_label_sample_i]['state'] = 'labeling'
    self.db['samples'][waiting_label_sample_i]['created_time'] = now_time
    self.db['samples'][waiting_label_sample_i]['assigner'] = user['full_name']

    # copy
    sample = copy.deepcopy(self.db['samples'][waiting_label_sample_i])

    # 转换时间到可显示
    for key in sample.keys():
      if key.endswith('time'):
        sample[key] = self.timestamp_2_str(sample[key])

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'sample': sample,
      'state': self.db['running']['state'],
      'stage': self.db['running']['stage'],
      'round': self.db['running']['round']
    })


class LoginHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    user_name = self.get_argument('user_name', None)
    user_password = self.get_argument('user_password', None)
    if user_name is None or user_password is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    white_users = self.db['white_users']
    if white_users is None:
      # 无需登录
      self.response(RESPONSE_STATUS_CODE.SUCCESS)
      return

    if user_name not in white_users:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    if user_password != self.db['white_users'][user_name]['password']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    cookie_id = str(uuid.uuid4())
    self.db['users'].update({
      cookie_id: {
        "full_name": user_name,
        'short_name': user_name[0].upper(),
        'labeling_sample': -1,
        'start_time': -1
      }
    })
    self.set_login_cookie({'cookie_id': cookie_id})
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class LogoutHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    self.clear_cookie('antgo')


class UserInfoHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 获取当前用户名
    user = self.get_current_user()
    if user is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # 遍历所有样本获得本轮，当前用户的标注记录
    statistic_info = {
      'sample_num': 0,
      'average_time': 0
    }

    for sample in self.db['samples']:
      for operator in sample['operators']:
        if operator['full_name'] == user['full_name']:
          statistic_info['sample_num'] += 1
          statistic_info['average_time'] += (sample['completed_time'] - sample['completed_time'])

    if statistic_info['sample_num'] != 0:
      statistic_info['average_time'] /= statistic_info['sample_num']

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={
      'user_name': user['full_name'],
      'short_name': user['short_name'],
      'task_name': self.db['task_metas']['task_name'],
      'task_type': self.db['task_metas']['task_type'],
      'project_type': 'LABEL',
      'statistic_info': statistic_info
    })


class LabelExportHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    export_samples = []
    for sample_id, sample in enumerate(self.db['samples']):
      export_sample = {
        'id': sample_id,
        'file_upload': sample['image_file'] if 'file_upload' not in sample else sample['file_upload'],
        'data': {'image': sample['image_file']},
        'created_at': self.timestamp_2_str(sample['created_time']),
        'updated_at': self.timestamp_2_str(sample['update_time']),
        'project': self.db['task_metas']['task_name'],
        'annotations': []
      }

      for sample_operator_i, sample_operator in enumerate(sample['operators']):
        temp = {
          'id': sample_operator_i,
          'completed_by': sample_operator['full_name'],
          'updated_at': self.timestamp_2_str(sample_operator['update_time']),
          'created_at': self.timestamp_2_str(sample_operator['create_time']),
          'result': []
        }

        for label_result in sample_operator['label_info']:
          if label_result['type'] in ['RECT', 'POINT', 'POLYGON']:
            v = copy.deepcopy(label_result['shape'])
            v['labels'] = label_result['label']
            temp['result'].append({
              'value': v,
              'width': sample['width'],
              'height': sample['height'],           
              'type': label_result['type']
            })
          elif label_result['type'] == 'CHOICES':
            v = {}
            v['labels'] = label_result['label']
            temp['result'].append({
              'value': v,
              'width': sample['width'],
              'height': sample['height'],
              'type': label_result['type']
            })

        export_sample['annotations'].append(temp)

      export_samples.append(export_sample)

    # 导出数据下载
    project_name = self.db['task_metas']['task_name']
    now_time = time.strftime("%Y-%m-%dx%H:%M:%S", time.localtime(time.time()))
    random_id = str(uuid.uuid4())
    export_file_name = f'{project_name}-{now_time}-{random_id}.json'
    with open(os.path.join(self.settings.get('static_path'), export_file_name), 'w') as fp:
      json.dump(export_samples, fp)

    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', f'attachment; filename={export_file_name}')
    with open(os.path.join(self.settings.get('static_path'), export_file_name), 'rb') as f:
      while True:
        data = f.read(MB)
        if not data:
            break
        self.write(data)

    self.finish()


class LabelImportHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    pass

class PingApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def label_server_start(
    dump_dir,
    server_port,
    task_metas,
    sample_metas,
    label_metas,
    sample_folder=None,
    sample_list=[],
    running_metas=None,
    white_users=None):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    static_dir = os.path.join(dump_dir, 'static')
    if not os.path.exists(static_dir):
      os.makedirs(static_dir)

    # 2.step launch show server
    db = {}
    # 添加默认样本列表
    db['samples'] = []

    if len(sample_list) > 0 and sample_folder is not None:
      sample_folder = os.path.abspath(sample_folder)  # 将传入的路径转换为绝对路径    
      # 为样本所在目录建立软连接到static下面
      os.system(f'cd {static_dir}; ln -s {sample_folder} dataset;')
                  
      # 将默认导入的样本直转换成新格式,并写入本地数据库
      for sample_i, sample_info in enumerate(sample_list):
        label_info = []
        # 添加目标框标注信息
        if len(sample_info['bboxes']) > 0:
          box_num = len(sample_info['bboxes'])
          for bi in range(box_num):
            box = sample_info['bboxes'][bi]
            label = sample_info['labels'][bi]
            label_info.append({
                  'type': 'RECT',
                  'shape': {
                    'x': box[0],
                    'y': box[1],
                    'width': box[2]-box[0],
                    'height': box[3]-box[1]
                  },
                  'label': label
            })
        elif 'bboxes' in sample_info['predictions']:
          box_num = len(sample_info['predictions']['bboxes'])
          for bi in range(box_num):
            box = sample_info['predictions']['bboxes'][bi]
            label = sample_info['predictions']['labels'][bi]
            label_info.append({
                  'type': 'RECT',
                  'shape': {
                    'x': box[0],
                    'y': box[1],
                    'width': box[2]-box[0],
                    'height': box[3]-box[1]
                  },
                  'label': label
            })
        
        # 添加多边形标注信息
        if len(sample_info['segments']) > 0:
          segment_num = len(sample_info['segments'])
          for bi in range(segment_num):
            points = sample_info['segments'][bi]
            label = sample_info['labels'][bi]
            label_info.append({
                  'type': 'POLYGON',
                  'shape': {
                    'points': points
                  },
                  'label': label
            })
        elif 'segments' in sample_info['predictions']:
          box_num = len(sample_info['predictions']['segments'])
          for bi in range(box_num):
            points = sample_info['predictions']['segments'][bi]
            label = sample_info['predictions']['segments'][bi]
            label_info.append({
                  'type': 'POLYGON',
                  'shape': {
                    'points': points
                  },
                  'label': label
            })
        
        # 添加关键点标注信息
        # pass
        
        # 添加图片类别
        if sample_info['image_label_name'] != '':
          label_info.append({
              'type': 'CHOICES',
              'choices': [sample_info['image_label_name']]
          })     
        elif 'image_label_name' in sample_info['predictions']:
          label_info.append({
              'type': 'CHOICES',
              'choices': [sample_info['predictions']['image_label_name']]
          })          
        
        # 加入数据库中
        db['samples'].append({
              'image_file': f'/static/dataset/{sample_info["image_file"]}' if sample_info['image_file'] != '' else sample_info['image_url'],
              # 'image_file': 'https://www.ssfiction.com/wp-content/uploads/2020/08/20200805_5f2b1669e9a24.jpg',
              'height': sample_info['height'],
              'width': sample_info['width'],
              'sample_id': sample_i,
              'completed_time': 0,
              'created_time': 0,
              'update_time': 0,
              'operators': [],
              'assigner': '',
              'state': 'waiting',
              'label_info': label_info,
        })
    
    # 设置样本基本信息
    db['sample_metas'] = sample_metas

    # 设置任务基本信息
    assert('task_type' in task_metas and 'task_name' in task_metas and 'label_type' in task_metas)
    db['task_metas'] = task_metas

    # 设置标注基本信息
    assert('category' in label_metas)
    label_category = label_metas['category']
    label_metas['label_category'] = label_category
    for i, label_config in enumerate(label_metas['label_category']):      
      if 'color' not in label_config or 'background_color' not in label_config:
        # 自动选择配色
        color = colormap.highlight[i%len(colormap.highlight)]['color']
        background_color = colormap.dark[i%len(colormap.dark)]['color']
        
        label_config['color'] = color
        label_config['background_color'] = background_color

    db['label_metas'] = label_metas

    # 设置白盒用户
    db['white_users'] = white_users

    # 其他初始化
    db['users'] = {}

    # 设置运行初始状态
    # state: 'running'/'abnormal'/'overtime'/'pending'
    # stage: 'waiting'/'training'/'labeling'/'finish'
    db['running'] = {}
    db['running']['state'] = 'running' if running_metas is None else running_metas['state']
    db['running']['stage'] = 'waiting' if running_metas is None else running_metas['stage']
    db['running']['round'] = 0
    db['running']['waiting_time_to_next_round'] = 0

    # cookie
    cookie_secret = base64.b64encode(uuid.uuid4().bytes + uuid.uuid4().bytes)

    settings = {
      'static_path': static_dir,
      'db': db,
      'cookie_secret': cookie_secret,
      'cookie_max_age_days': 30,
      'Content-Security-Policy': "frame-ancestors 'self' {}".format('http://localhost:8080/')
    }

    app = tornado.web.Application(
      handlers=[
        (r"/antgo/api/label/task/", LabelTaskHandler),                    # 获得系统任务
        (r"/antgo/api/label/sample/", LabelSamplesHandler),               # 获得所有样本（TODO, 支持lazyload）
        (r"/antgo/api/label/sample/fresh/", LabelFreshSampleHandler),     # 更新所有样本
        (r"/antgo/api/label/sample/update/", LabelUpdateSampleHandler),   # 更新样本标注结果
        (r"/antgo/api/label/sample/get/", LabelGetSampleHandler),         # 得到指定样本信息
        (r"/antgo/api/label/sample/next/", LabelNextSampleHandler),       # 得到下一个需要进行标注的样本信息
        (r"/antgo/api/label/export/", LabelExportHandler),                # 标注导出
        (r"/antgo/api/label/export/download/", LabelExportHandler),       # 标注导出
        (r"/antgo/api/label/import/", LabelImportHandler),                # 标注导入
        (r"/antgo/api/user/info/", UserInfoHandler),                  # 获得用户信息
        (r"/antgo/api/user/login/", LoginHandler),  # 登录，仅支持预先指定用户
        (r"/antgo/api/user/logout/", LogoutHandler),  # 退出
        (r"/antgo/api/ping/", PingApiHandler),  # ping 服务
        (r"/antgo/api/info/", LabelStateHandler),  # 获得当前系统状态
        (r'/(.*)', tornado.web.StaticFileHandler, {"path": static_dir, "default_filename": "index.html"})
      ],
      **settings
    )
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('ANTGO-LABEL is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('ANTGO-LABEL stop server')
  except GracefulExitException:
    logger.info('ANTGO-LABEL server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass


if __name__ == '__main__':
  task_metas = {
      'task_type': 'OBJECT-DET',      #
      'task_name': 'HELLO',
      # 'label_type': 'RECT',
      # 'label_type': 'POINT',
      'label_type': 'POLYGON'
    }
  sample_metas = {
    'filters': ['id', 'time']
  }

  running_metas = {
    'state': 'running',
    'stage': 'labeling'
  }
  label_metas = {
        'category': [
          {
            'class_name': 'A',
            'class_index': 0,
            'color': 'green',
            'background_color': '#00800026'
          },
          {
            'class_name': 'B',
            'class_index': 1,
            'color': 'blue',
            'background_color': '#0000ff26'
          },
        ],
      }
  white_users = {
    'jian@baidu.com': {
      'password': '112233'
    }
  }
  samples = [
            {
              'image_file': '/static/data/1.png',
              'height': 800,
              'width': 1200,
              'sample_id': 0,
              'completed_time': 0,
              'created_time': 0,
              'update_time': 0,
              'operators': [],
              'assigner': '',
              'state': 'waiting',
              'label_info': [
                {
                  'type': 'RECT',
                  'shape': {
                    'x': 0,
                    'y': 0,
                    'width': 100,
                    'height': 100
                  },
                  'label': 0
                },
                {
                  'type': 'RECT',
                  'shape': {
                    'x': 100,
                    'y': 100,
                    'width': 100,
                    'height': 100
                  },
                  'label': 1
                }
              ],
            },
            {
              'image_file': '/static/data/1.png',
              'height': 800,
              'width': 1200,
              'sample_id': 1,
              'completed_time': 0,
              'created_time': 0,
              'update_time': 0,
              'operators': [],
              'state': 'waiting',
              'assigner': '',
              'label_info': [],
            },
            {
              'image_file': '/static/data/1.png',
              'height': 800,
              'width': 1200,
              'sample_id': 2,
              'completed_time': 0,
              'created_time': 0,
              'update_time': 0,
              'state': 'waiting',
              'assigner': '',
              'operators': [],
              'label_info': [],
            }
          ] * 100
  for i in range(len(samples)):
    samples[i] = copy.deepcopy(samples[i])
    samples[i]['sample_id'] = i

  label_server_start('/Users/bytedance/Downloads/workspace/my/A', 9000, task_metas, sample_metas, label_metas, samples, running_metas, white_users)



# -*- coding: UTF-8 -*-
# @Time    : 2019/1/23 3:13 PM
# @File    : sampling_def.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

"""Abstract class for sampling methods.

Provides interface to sampling methods that allow same signature
for select_batch.  Each subclass implements select_batch_ with the desired
signature for readability.
"""

import abc
import numpy as np

class SamplingMethod(object):
  __metaclass__ = abc.ABCMeta

  def flatten_X(self, X):
    shape = X.shape
    flat_X = X
    if len(shape) > 2:
      flat_X = np.reshape(X, (shape[0], np.product(shape[1:])))
    return flat_X


  @abc.abstractmethod
  def select_batch_(self):
    return

  def select_batch(self, **kwargs):
    return self.select_batch_(**kwargs)

  def to_dict(self):
    return None


# 优化器配置
optimizer = dict(type='SGD', lr=0.01,  weight_decay=5e-4, momentum=0.01, nesterov=True)
optimizer_config = dict(grad_clip=dict(max_norm=10))

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
)

# 日志配置
log_config = dict(
    interval=1,    
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 模型配置
model = dict(
    type='TTFNet',
    backbone=dict(
        type='ResNet',
            depth=50, 
            in_channels=3, 
            out_indices=[3,2,1]),      
    neck=dict(type="FPN", in_channels=[512, 1024, 2048], out_channels=32, num_outs=1),
    bbox_head=dict(
        type='FcosHead',
        in_channel=32,
        feat_channel=32,
        num_classes=4,
        down_stride=4,
        score_thresh=0.2,
        train_cfg=None,
        test_cfg=dict(topk=100, local_maximum_kernel=3),
        loss_ch=dict(type='GaussianFocalLoss', loss_weight=1.0)
    ),  
)

# 描述模型基本信息
info = dict(
    backbone=dict(
        channels=[512, 1024, 2048],
        shapes=[32,16,8]
    )
)

# checkpoint配置
checkpoint_config = dict(interval=1, out_dir='./output/')       

# 数据配置
data=dict(
    train=dict(
        type='VisalSO',
        train_or_test='train',
        dir='./visalso_dataset',
        pipeline=[
            dict(type="ResizeByShort", short_size=256),
            dict(type="RandomScaledCrop", target_size=(256,256)),
            dict(type="Rotation", degree=15),
            dict(type='ResizeS', target_dim=[256,256]),
            dict(type='ColorDistort', hue=[-5,5,0.3], saturation=[0.8,1.2,0.3], contrast=[0.8,1.2,0.3], brightness=[0.8,1.2,0.3]),
            dict(type='RandomFlipImage', swap_labels=[]),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='IImageToTensor', keys=['image'])
        ],
        inputs_def=dict(
            fields = ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    train_dataloader=dict(
        samples_per_gpu=4, 
        workers_per_gpu=1,
        drop_last=True,
        shuffle=True,
        ignore_stack=['bboxes', 'labels', 'image_meta']
    ),
    val=dict(
        type='VisalSO',
        train_or_test='val',
        dir='./visalso_dataset',
        pipeline=[
            dict(type="ResizeByShort", short_size=256),
            dict(type='FixedCrop', target_size=(256,256), align='center'),
            dict(type='ResizeS', target_dim=[256,256]),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='IImageToTensor', keys=['image'])
        ],
        inputs_def=dict(
            fields= ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    val_dataloader=dict(
        samples_per_gpu=4, 
        workers_per_gpu=1,
        drop_last=False,
        shuffle=False,
        ignore_stack=['bboxes', 'labels', 'image_meta']
    ),
    test=dict(
        type='VisalSO',
        train_or_test='test',
        dir='./visalso_dataset',
        pipeline=[
            dict(type="ResizeByShort", short_size=256),
            dict(type='FixedCrop', target_size=(256,256), align='center'),
            dict(type='ResizeS', target_dim=[256,256]),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='IImageToTensor', keys=['image'])
        ],
        inputs_def=dict(
            fields= ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    test_dataloader=dict(
        samples_per_gpu=4, 
        workers_per_gpu=1,
        drop_last=False,
        shuffle=False,
        ignore_stack=['bboxes', 'labels', 'image_meta']
    )
)

# 评估方案配置
evaluation=dict(
    out_dir='./output/', 
    interval=1, 
    metric=dict(
        type='COCOCompatibleEval', 
        categories=[{'name': f'{label}', 'id': label} for label in range(4)]
    ), 
    save_best='AP@[ IoU=0.50:0.95 | area= all | maxDets=100 ]',
    rule='greater'
)

# 导出配置
export=dict(
    input_shape_list = [[1,3,256,256]],
    input_name_list=["image"],
    output_name_list=["heatmap", "offset"]
)

max_epochs = 60


# -*- coding: UTF-8 -*-
# @Time    : 2019/1/22 1:16 PM
# @File    : ensemble.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from random import shuffle
from antgo.resource.html import *
from antgo.ant.base import *
from antgo.dataflow.common import *
from antgo.dataflow.recorder import *
from antvis.client.httprpc import *
from antgo.task.task import *
import traceback
import subprocess
import os
import socket
import json
import zipfile
import multiprocessing
from antgo.crowdsource.ensemble_server import *
from antgo.dataflow.dataset.proxy_dataset import *
import requests
from io import BytesIO
import numpy as np
import pickle


class EnsembleMergeRecorder(object):
    def __init__(self, role='master',
                 ip='', port='',
                 experiment_uuid='',
                 dump_dir='',
                 ensemble_method='',
                 dataset_name='',
                 dataset_flag='train',
                 model_weight=1.0,
                 prefix='',
                 feedback=True):
        self.role = role        # master, worker
        self._dump_dir = ''
        self.url = 'http://%s:%s'%(ip, (str)(port))
        self.experiment_uuid = experiment_uuid
        self.record_variable_name = []
        self.record_sample_list = {}
        self.dataset_name = dataset_name
        self.dataset_flag = dataset_flag
        self.ensemble_method = ensemble_method
        self.dump_dir = dump_dir
        self.model_weight = model_weight
        self.feedback = feedback
        self.prefix = prefix
        self.worker_number = f'{uuid.uuid4()}'
        
        self.data_uuid = 0          # 用于数据分发
        self.avg_data_uuid = 0      # 用于数据聚合

    # 等待数据ready后，返回
    def get(self):
        proxies = {
                "http": None,
                'https': None
            }

        try_count = 0
        while True:
            response = \
                requests.get('%s/antgo/api/ensemble/get/'%(self.url),
                                headers={'id': f'{self.data_uuid}'},
                                proxies=proxies, stream=True)

            if response.status_code == 200:
                with BytesIO() as pdf:
                    for chunk in response.iter_content(chunk_size=1024*1024):
                        if chunk:
                            pdf.write(chunk)
                    
                    data = pickle.loads(pdf.getvalue())
                
                # 索引编号++
                self.data_uuid += 1
                return data

            logging.warn(f'Get data ID-{self.data_uuid} overtime.')
            try_count += 1
            
            if try_count > 5:
                break
                
        self.data_uuid += 1
        # TODO, 对于失败获取数据的情况，外部应该直接跳此样本
        # 如何补救留给外部调用者
        # 即使当前数据失败，索引编号++
        return None
    
    # 推送成功后，立即返回
    def put(self, data):
        proxies = {
            "http": None,
            'https': None
        }

        # 将数据以文件流模式上传
        data = pickle.dumps(data)
        response = \
            requests.post('%s/antgo/api/ensemble/put/'%(self.url),
                            headers={'id': f'{self.data_uuid}'},
                            data=data,
                            proxies=proxies, stream=True)

        if response.status_code != 200:
            logging.error(f'Put data ID-{self.data_uuid} error.')
            self.data_uuid += 1
            return False
                        
        self.data_uuid += 1
        return True
    
    def avg(self, data=None):
        proxies = {
            "http": None,
            'https': None
        }

        if data is None:
            response = \
                requests.post('%s/antgo/api/ensemble/avg/'%(self.url),
                                headers={
                                    'file_id': f'{self.avg_data_uuid}',
                                    'id': f'{self.avg_data_uuid}',
                                    'feedback': f"{self.feedback}"
                                },
                                proxies=proxies)           

            self.avg_data_uuid += 1
            if response.status_code != 200:
                logging.warn(f'Get ensemble data erro.')
                # TODO, 对于失败获取数据的情况，外部应该直接跳此样本
                # 如何补救留给外部调用者
                return None

            with BytesIO() as pdf:
                for chunk in response.iter_content(chunk_size=1024*1024):
                    if chunk:
                        pdf.write(chunk)
                data = pickle.loads(pdf.getvalue())
            
            return data
        
        # check data type
        for _, v in data.items():
            if not isinstance(v, np.ndarray):
                logging.error("Has non np.ndarray data.")
                return None
        
        # start ensemble
        data = pickle.dumps(data)
        response = \
            requests.post('%s/antgo/api/ensemble/avg/'%(self.url),
                            headers={
                                'file_id': f'{self.avg_data_uuid}',
                                'id': f'{self.avg_data_uuid}',
                                'worker_prefix': f'{self.prefix}',
                                'weight': f"{self.model_weight}",
                                'feedback': f"{self.feedback}"
                            },
                            data=data,
                            proxies=proxies, stream=True)

        self.avg_data_uuid += 1
        if response.status_code != 200:
            logging.warn(f'Get ensemble data erro.')
            return
    
        if self.feedback:
            with BytesIO() as pdf:
                for chunk in response.iter_content(chunk_size=1024*1024):
                    if chunk:
                        pdf.write(chunk)
                data = pickle.loads(pdf.getvalue())
                return data
        return None

    def record(self, data):
        assert('id' in data or 'image_file' in data)
        id = ''
        if 'id' in data:
            id = data['id']
            data.pop('id')
        if 'image_file' in data:
            id = data['image_file']
            data.pop('image_file')

        # 记录样本ID
        if id not in self.record_sample_list:
            self.record_sample_list[id] = {}

        # 记录样本记录的数据
        for k in data.keys():
            if k not in self.record_variable_name:
                self.record_variable_name.append(k)
            if k not in self.record_sample_list[id]:
                self.record_sample_list[id][k] = 0

            self.record_sample_list[id][k] += 1

        # 记录样本
        for k, v in data.items():
            name = k
            value = v['data']
            if type(value) != np.ndarray:
                logger.warn('Record %s only support numpy array.'%name)
                continue

            # 本地保存数据
            key = '%s.%s.%s.data'%(self.experiment_uuid, id, name)
            # if not os.path.exists(os.path.join(self.dump_dir, '%s/%s'%(self.experiment_uuid, id))):
            #     os.makedirs(os.path.join(self.dump_dir, '%s/%s'%(self.experiment_uuid, id)))

            with open(os.path.join(self.dump_dir, key), 'wb') as fp:
                fp.write(msgpack.packb(value, default=ms.encode))

    def close(self):
        if len(self.record_sample_list) == 0:
            return

        # 提交完成 ensemble 模型创建
        record_sample_num = len(self.record_sample_list)

        # 记录信息（MLTALKER）
        logger.info('Update ensemble experiment record.')
        info = {
            'record_sample_num': record_sample_num,
            'dataset_name': self.dataset_name,
            'dataset_flag': self.dataset_flag,
            'ensemble_method': self.ensemble_method,
            'record_variable_name': self.record_variable_name,
            'ensemble_url': self.url            # 记录在平台，聚合数据URL
        }

        response = mlogger.info.ensemble.merge.post(info=json.dumps(info))
        if response['status'] == 'OK':
            logger.info('Finish ensemble model submit.')
        else:
            logger.error('Fail ensemble model submit')

    @property
    def dump_dir(self):
        return self._dump_dir

    @dump_dir.setter
    def dump_dir(self, val):
        self._dump_dir = val


class EnsembleReleaseRecorder(object):
    def __init__(self, ensemble_uuid):
        # 聚合名字
        self.ensemble_uuid = ensemble_uuid

        # 获得实验列表
        response = mlogger.info.ensemble.release.get(ensemble_uuid=ensemble_uuid)
        content = response['content']
        # [{'experiment_uuid': '', 'experiment_weight': 1.0, 'experiment_user': , 'experiment_url': },{}]
        self.ensemble_models = self.parse_ensemble_model(content)

        self._dump_dir = ''

    def recursive_parse(self, root, root_weight, root_developer, data):
        if 'experiment_uuid' in root:
            for model, weight, developer, server_url in \
                zip(root['experiment_uuid'], root['experiment_weight'], root['experiment_user'],
                    root['experiment_url']):
                data.append({
                    'experiment_uuid': model,
                    'experiment_weight': (float)(weight) * (float)(root_weight),
                    'experiment_user': developer,
                    'experiment_url': server_url
                })
            return

        for model, weight, developer in zip(root['ensemble_uuid'], root['ensemble_weight'], root['ensemble_user']):
            self.recursive_parse(model, (float)(weight), developer, data)

    def parse_ensemble_model(self, content):
        # {'experiment_uuid': [], 'experiment_weight': []}
        # {'ensemble_uuid': [{'ensemble_uuid':[], 'ensemble_weight': []}], 'ensemble_weight': [1.0]}
        if 'experiment_uuid' in content:
            data = []
            for model, weight, developer, serve_url in \
                zip(content['experiment_uuid'], content['experiment_weight'], content['experiment_user'],
                    content['experiment_url']):
                data.append({
                    'experiment_uuid': model,
                    'experiment_weight': (float)(weight),
                    'experiment_user': developer,
                    'experiment_url': serve_url
                })

            return data

        data = []
        for model, weight, developer in \
            zip(content['ensemble_uuid'], content['ensemble_weight'], content['ensemble_user']):
            self.recursive_parse(model, (float)(weight), developer, data)
        return data

    def get(self, kwargs):
        assert ('id' in kwargs)
        id = str(kwargs['id']['data'])
        kwargs.pop('id')

        result = {}
        for k, v in kwargs.items():
            name = k

            merge_data = 0.0
            merge_weight = 0.0
            for ensemble_mode in self.ensemble_models:
                experiment_uuid = ensemble_mode['experiment_uuid']
                experiment_url = ensemble_mode['experiment_url']
                model_developer = ensemble_mode['experiment_user']
                model_weight = ensemble_mode['experiment_weight']

                key = '%s.%s.%s.data' % (experiment_uuid, id, name)
                data = mlogger.file.download(None,
                                             key,
                                             None,
                                             f'{experiment_url}/ensemble-api/data/')
                if data is None:
                    continue
                data = msgpack.unpackb(data, object_hook=ms.decode)

                merge_data += data * model_weight
                merge_weight += model_weight

            result[k] = {
                'data': merge_data / merge_weight
            }

        return result

    def record(self, data):
        pass

    def close(self):
        pass

    @property
    def dump_dir(self):
        return self._dump_dir

    @dump_dir.setter
    def dump_dir(self, val):
        self._dump_dir = val


class AntEnsemble(AntBase):
    def __init__(self,
                 ant_context,
                 ant_name,
                 ant_data_folder,
                 ant_dump_dir,
                 ant_token,
                 ant_task_config,
                 dataset,
                 **kwargs):
        super().__init__(ant_name, ant_context=ant_context, ant_token=ant_token, **kwargs)
        # master, slave
        # bagging, stacking
        self.ensemble_stage =  self.context.params.ensemble.stage            # train, merge, release
        assert(self.ensemble_stage in ['train', 'merge', 'release'])

        self.mode = self.context.params.ensemble.mode      # online/offline
        self.method = self.context.params.ensemble.method   # bagging,stacking,blending,bosting
        self.ant_dump_dir = ant_dump_dir

        self.ant_task_config = ant_task_config
        self.ant_data_source = ant_data_folder
        self.dataset = dataset
        self.model_weight = 1.0

    def start(self):
        # 1.step 加载训练任务
        running_ant_task = None
        if self.token is not None:
            # 1.1.step load train task
            response = mlogger.info.challenge.get(command=type(self).__name__)
            if response['status'] == 'ERROR':
                # invalid token
                logger.error('Couldnt load train task.')
                self.token = None
            elif response['status'] in ['OK', 'SUSPEND']:
                content = response['content']

                # maybe user token or task token
                if 'task' in content:
                    # task token
                    challenge_task = create_task_from_json(content)
                    if challenge_task is None:
                        logger.error('Couldnt load train task.')
                        exit(-1)
                    running_ant_task = challenge_task
            else:
                # unknow error
                logger.error('Unknow error.')
                exit(-1)

        if running_ant_task is None:
            # 1.2.step load custom task
            if self.ant_task_config is not None:
                custom_task = create_task_from_xml(self.ant_task_config, self.context)
                if custom_task is None:
                    logger.error('Couldnt load custom task.')
                    exit(-1)
                running_ant_task = custom_task

        assert (running_ant_task is not None)

        # 启动服务器
        background_server_process = None
        if self.ensemble_stage == 'merge' and self.context.params.ensemble.role == 'master':
            worker_num = (int)(self.context.params.ensemble.worker)
            server_ip = self.context.params.system.ip
            server_port = (int)(self.context.params.system.port)
            background_server_process = \
                multiprocessing.Process(target=ensemble_server_start,
                                        args=(self.ant_dump_dir,
                                              server_port,
                                              worker_num,
                                              self.context.params.ensemble.get('uncertain_vote', None)))
            if not self.context.params.ensemble.get('background', True):
                # 后台服务
                background_server_process.daemon = True
            background_server_process.start()

        # 等待master服务开启
        while self.ensemble_stage == 'merge':
            try:
                server_ip = self.context.params.system.ip
                server_port = (int)(self.context.params.system.port)
                proxies = {
                    "http": None,
                    'https': None
                }
                result = requests.get('http://%s:%d/antgo/api/ping/'%(server_ip, server_port), proxies=proxies)
                if result.status_code == 200:
                    break
            except:
                logger.info("Waiting ensemble server launch.")
                time.sleep(10)

        # 数据集
        dataset_flag = 'train'
        if self.ensemble_stage == 'train':
            dataset_flag = 'train'

        dataset_name = ''
        if self.dataset is not None and self.dataset != '':
            if len(self.dataset.split('/')) == 2:
                dataset_name, dataset_flag = self.dataset.split('/')
            else:
                dataset_name = self.dataset
                if self.ensemble_stage in ['merge', 'release']:
                    dataset_flag = 'test'
                else:
                    dataset_flag = 'train'
        elif 'dataset' in self.context.params.ensemble.keys():
            if len(self.context.params.ensemble.dataset.split('/')) == 2:
                dataset_name, dataset_flag = self.context.params.ensemble.dataset.split('/')
            else:
                dataset_name = self.context.params.ensemble.dataset
                if self.ensemble_stage in ['merge', 'release']:
                    dataset_flag = 'test'
                else:
                    dataset_flag = 'train'
        else:
            dataset_name = running_ant_task.dataset_name
            if len(dataset_name.split('/')) == 2:
                dataset_name = dataset_name.split('/')[0]

            if self.ensemble_stage in ['merge', 'release']:
                dataset_flag = 'test'
            else:
                dataset_flag = 'train'

        # 模型权重
        if 'weight' in self.context.params.ensemble.keys():
            self.model_weight = self.context.params.ensemble.weight

        logger.info('Using dataset %s/%s.'%(dataset_name, dataset_flag))

        ant_dataset = None
        if self.context.register_at(dataset_flag) is not None:
            ant_dataset = ProxyDataset(dataset_flag)
            kwargs = {dataset_flag: self.context.register_at(dataset_flag)}
            ant_dataset.register(**kwargs)
        elif running_ant_task.dataset is not None:
            ant_dataset = \
                running_ant_task.dataset(dataset_flag,
                                         os.path.join(self.ant_data_source, dataset_name),
                                         running_ant_task.dataset_params)

        self.stage = 'ENSEMBLE-%s' % self.ensemble_stage
        if self.ensemble_stage == 'merge':
            dump_dir = os.path.join(self.ant_dump_dir, 'ensemble', 'merge')
            if os.path.exists(dump_dir):
                shutil.rmtree(dump_dir)
            if not os.path.exists(dump_dir):
                os.makedirs(dump_dir)

            logger.info('Start ensemble merge process.')
            self.context.recorder = \
                EnsembleMergeRecorder(role=self.context.params.ensemble.role,
                                      ip=self.context.params.system.ip,
                                      port=self.context.params.system.port,
                                      experiment_uuid=self.experiment_uuid,
                                      dump_dir=dump_dir,
                                      ensemble_method=self.method,
                                      dataset_name=dataset_name,
                                      dataset_flag=dataset_flag,
                                      prefix=f"{self.ant_name}-{self.context.params.ensemble.get('model_name', 'model')}",
                                      model_weight=self.model_weight,
                                      feedback=self.context.params.ensemble.feedback)

            if self.context.is_interact_mode:
                if self.context.params.ensemble.get('background', True):
                    self.context.registry_clear_callback(lambda : background_server_process.join())
                return

            with safe_recorder_manager(self.context.recorder):
                try:
                    self.context.call_infer_process(ant_dataset, dump_dir)
                except Exception as e:
                    if type(e.__cause__) != StopIteration:
                        print(e)
                        traceback.print_exc()

            if self.context.params.ensemble.get('background', True):
                background_server_process.join()

        elif self.ensemble_stage == 'release':
            dump_dir = os.path.join(self.ant_dump_dir, 'ensemble', 'release')
            if not os.path.exists(dump_dir):
                os.makedirs(dump_dir)

            logger.info('Start ensemble release process.')
            ensemble_uuid = self.context.params.ensemble.get('uuid', '')
            self.context.recorder = EnsembleReleaseRecorder(ensemble_uuid)

            if self.context.is_interact_mode:
                return

            with safe_recorder_manager(self.context.recorder):
                try:
                    self.context.call_infer_process(ant_dataset, dump_dir)
                except Exception as e:
                    if type(e.__cause__) != StopIteration:
                        print(e)
                        traceback.print_exc()
        else:
            # 数据重采样
            dump_dir = os.path.join(self.ant_dump_dir, 'ensemble', 'train')
            if not os.path.exists(dump_dir):
                os.makedirs(dump_dir)

            logger.info('Start ensemble train process.')
            if self.context.is_interact_mode:
                return

            try:
                self.context.call_training_process(ant_dataset, dump_dir)
            except Exception as e:
                if type(e.__cause__) != StopIteration:
                    print(e)
                    traceback.print_exc()

